# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

from ccxt.base.exchange import Exchange
from ccxt.base.errors import ExchangeError
from ccxt.base.errors import AuthenticationError
from ccxt.base.errors import PermissionDenied
from ccxt.base.errors import InsufficientFunds
from ccxt.base.errors import InvalidOrder
from ccxt.base.errors import OrderNotFound
from ccxt.base.errors import InvalidNonce
from ccxt.base.decimal_to_precision import TICK_SIZE


class bitbank(Exchange):

    def describe(self):
        return self.deep_extend(super(bitbank, self).describe(), {
            'id': 'bitbank',
            'name': 'bitbank',
            'countries': ['JP'],
            'version': 'v1',
            'has': {
                'CORS': None,
                'spot': True,
                'margin': False,
                'swap': False,
                'future': False,
                'option': False,
                'addMargin': False,
                'cancelOrder': True,
                'createOrder': True,
                'createReduceOnlyOrder': False,
                'fetchBalance': True,
                'fetchBorrowRate': False,
                'fetchBorrowRateHistories': False,
                'fetchBorrowRateHistory': False,
                'fetchBorrowRates': False,
                'fetchBorrowRatesPerSymbol': False,
                'fetchDepositAddress': True,
                'fetchFundingHistory': False,
                'fetchFundingRate': False,
                'fetchFundingRateHistory': False,
                'fetchFundingRates': False,
                'fetchIndexOHLCV': False,
                'fetchLeverage': False,
                'fetchLeverageTiers': False,
                'fetchMarginMode': False,
                'fetchMarkOHLCV': False,
                'fetchMyTrades': True,
                'fetchOHLCV': True,
                'fetchOpenInterestHistory': False,
                'fetchOpenOrders': True,
                'fetchOrder': True,
                'fetchOrderBook': True,
                'fetchPosition': False,
                'fetchPositionMode': False,
                'fetchPositions': False,
                'fetchPositionsRisk': False,
                'fetchPremiumIndexOHLCV': False,
                'fetchTicker': True,
                'fetchTrades': True,
                'fetchTradingFee': False,
                'fetchTradingFees': True,
                'fetchTransfer': False,
                'fetchTransfers': False,
                'reduceMargin': False,
                'setLeverage': False,
                'setMarginMode': False,
                'setPositionMode': False,
                'transfer': False,
                'withdraw': True,
            },
            'timeframes': {
                '1m': '1min',
                '5m': '5min',
                '15m': '15min',
                '30m': '30min',
                '1h': '1hour',
                '4h': '4hour',
                '8h': '8hour',
                '12h': '12hour',
                '1d': '1day',
                '1w': '1week',
            },
            'hostname': 'bitbank.cc',
            'urls': {
                'logo': 'https://user-images.githubusercontent.com/1294454/37808081-b87f2d9c-2e59-11e8-894d-c1900b7584fe.jpg',
                'api': {
                    'public': 'https://public.{hostname}',
                    'private': 'https://api.{hostname}',
                    'markets': 'https://api.{hostname}',
                },
                'www': 'https://bitbank.cc/',
                'doc': 'https://docs.bitbank.cc/',
                'fees': 'https://bitbank.cc/docs/fees/',
            },
            'api': {
                'public': {
                    'get': [
                        '{pair}/ticker',
                        '{pair}/depth',
                        '{pair}/transactions',
                        '{pair}/transactions/{yyyymmdd}',
                        '{pair}/candlestick/{candletype}/{yyyymmdd}',
                    ],
                },
                'private': {
                    'get': [
                        'user/assets',
                        'user/spot/order',
                        'user/spot/active_orders',
                        'user/spot/trade_history',
                        'user/withdrawal_account',
                    ],
                    'post': [
                        'user/spot/order',
                        'user/spot/cancel_order',
                        'user/spot/cancel_orders',
                        'user/spot/orders_info',
                        'user/request_withdrawal',
                    ],
                },
                'markets': {
                    'get': [
                        'spot/pairs',
                    ],
                },
            },
            'precisionMode': TICK_SIZE,
            'exceptions': {
                '20001': AuthenticationError,
                '20002': AuthenticationError,
                '20003': AuthenticationError,
                '20005': AuthenticationError,
                '20004': InvalidNonce,
                '40020': InvalidOrder,
                '40021': InvalidOrder,
                '40025': ExchangeError,
                '40013': OrderNotFound,
                '40014': OrderNotFound,
                '50008': PermissionDenied,
                '50009': OrderNotFound,
                '50010': OrderNotFound,
                '60001': InsufficientFunds,
                '60005': InvalidOrder,
            },
        })

    def fetch_markets(self, params={}):
        """
        retrieves data on all markets for bitbank
        :param dict params: extra parameters specific to the exchange api endpoint
        :returns [dict]: an array of objects representing market data
        """
        response = self.marketsGetSpotPairs(params)
        #
        #     {
        #       "success": 1,
        #       "data": {
        #         "pairs": [
        #           {
        #             "name": "btc_jpy",
        #             "base_asset": "btc",
        #             "quote_asset": "jpy",
        #             "maker_fee_rate_base": "0",
        #             "taker_fee_rate_base": "0",
        #             "maker_fee_rate_quote": "-0.0002",
        #             "taker_fee_rate_quote": "0.0012",
        #             "unit_amount": "0.0001",
        #             "limit_max_amount": "1000",
        #             "market_max_amount": "10",
        #             "market_allowance_rate": "0.2",
        #             "price_digits": 0,
        #             "amount_digits": 4,
        #             "is_enabled": True,
        #             "stop_order": False,
        #             "stop_order_and_cancel": False
        #           }
        #         ]
        #       }
        #     }
        #
        data = self.safe_value(response, 'data')
        pairs = self.safe_value(data, 'pairs', [])
        result = []
        for i in range(0, len(pairs)):
            entry = pairs[i]
            id = self.safe_string(entry, 'name')
            baseId = self.safe_string(entry, 'base_asset')
            quoteId = self.safe_string(entry, 'quote_asset')
            base = self.safe_currency_code(baseId)
            quote = self.safe_currency_code(quoteId)
            result.append({
                'id': id,
                'symbol': base + '/' + quote,
                'base': base,
                'quote': quote,
                'settle': None,
                'baseId': baseId,
                'quoteId': quoteId,
                'settleId': None,
                'type': 'spot',
                'spot': True,
                'margin': False,
                'swap': False,
                'future': False,
                'option': False,
                'active': self.safe_value(entry, 'is_enabled'),
                'contract': False,
                'linear': None,
                'inverse': None,
                'taker': self.safe_number(entry, 'taker_fee_rate_quote'),
                'maker': self.safe_number(entry, 'maker_fee_rate_quote'),
                'contractSize': None,
                'expiry': None,
                'expiryDatetime': None,
                'strike': None,
                'optionType': None,
                'precision': {
                    'amount': self.parse_number(self.parse_precision(self.safe_string(entry, 'amount_digits'))),
                    'price': self.parse_number(self.parse_precision(self.safe_string(entry, 'price_digits'))),
                },
                'limits': {
                    'leverage': {
                        'min': None,
                        'max': None,
                    },
                    'amount': {
                        'min': self.safe_number(entry, 'unit_amount'),
                        'max': self.safe_number(entry, 'limit_max_amount'),
                    },
                    'price': {
                        'min': None,
                        'max': None,
                    },
                    'cost': {
                        'min': None,
                        'max': None,
                    },
                },
                'info': entry,
            })
        return result

    def parse_ticker(self, ticker, market=None):
        symbol = self.safe_symbol(None, market)
        timestamp = self.safe_integer(ticker, 'timestamp')
        last = self.safe_string(ticker, 'last')
        return self.safe_ticker({
            'symbol': symbol,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'high': self.safe_string(ticker, 'high'),
            'low': self.safe_string(ticker, 'low'),
            'bid': self.safe_string(ticker, 'buy'),
            'bidVolume': None,
            'ask': self.safe_string(ticker, 'sell'),
            'askVolume': None,
            'vwap': None,
            'open': None,
            'close': last,
            'last': last,
            'previousClose': None,
            'change': None,
            'percentage': None,
            'average': None,
            'baseVolume': self.safe_string(ticker, 'vol'),
            'quoteVolume': None,
            'info': ticker,
        }, market)

    def fetch_ticker(self, symbol, params={}):
        """
        fetches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific market
        :param str symbol: unified symbol of the market to fetch the ticker for
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns dict: a `ticker structure <https://docs.ccxt.com/en/latest/manual.html#ticker-structure>`
        """
        self.load_markets()
        market = self.market(symbol)
        request = {
            'pair': market['id'],
        }
        response = self.publicGetPairTicker(self.extend(request, params))
        data = self.safe_value(response, 'data', {})
        return self.parse_ticker(data, market)

    def fetch_order_book(self, symbol, limit=None, params={}):
        """
        fetches information on open orders with bid(buy) and ask(sell) prices, volumes and other data
        :param str symbol: unified symbol of the market to fetch the order book for
        :param int|None limit: the maximum amount of order book entries to return
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/en/latest/manual.html#order-book-structure>` indexed by market symbols
        """
        self.load_markets()
        market = self.market(symbol)
        request = {
            'pair': market['id'],
        }
        response = self.publicGetPairDepth(self.extend(request, params))
        orderbook = self.safe_value(response, 'data', {})
        timestamp = self.safe_integer(orderbook, 'timestamp')
        return self.parse_order_book(orderbook, market['symbol'], timestamp)

    def parse_trade(self, trade, market=None):
        timestamp = self.safe_integer(trade, 'executed_at')
        market = self.safe_market(None, market)
        priceString = self.safe_string(trade, 'price')
        amountString = self.safe_string(trade, 'amount')
        id = self.safe_string_2(trade, 'transaction_id', 'trade_id')
        takerOrMaker = self.safe_string(trade, 'maker_taker')
        fee = None
        feeCostString = self.safe_string(trade, 'fee_amount_quote')
        if feeCostString is not None:
            fee = {
                'currency': market['quote'],
                'cost': feeCostString,
            }
        orderId = self.safe_string(trade, 'order_id')
        type = self.safe_string(trade, 'type')
        side = self.safe_string(trade, 'side')
        return self.safe_trade({
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'symbol': market['symbol'],
            'id': id,
            'order': orderId,
            'type': type,
            'side': side,
            'takerOrMaker': takerOrMaker,
            'price': priceString,
            'amount': amountString,
            'cost': None,
            'fee': fee,
            'info': trade,
        }, market)

    def fetch_trades(self, symbol, since=None, limit=None, params={}):
        """
        get the list of most recent trades for a particular symbol
        :param str symbol: unified symbol of the market to fetch trades for
        :param int|None since: timestamp in ms of the earliest trade to fetch
        :param int|None limit: the maximum amount of trades to fetch
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns [dict]: a list of `trade structures <https://docs.ccxt.com/en/latest/manual.html?#public-trades>`
        """
        self.load_markets()
        market = self.market(symbol)
        request = {
            'pair': market['id'],
        }
        response = self.publicGetPairTransactions(self.extend(request, params))
        data = self.safe_value(response, 'data', {})
        trades = self.safe_value(data, 'transactions', [])
        return self.parse_trades(trades, market, since, limit)

    def fetch_trading_fees(self, params={}):
        """
        fetch the trading fees for multiple markets
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns dict: a dictionary of `fee structures <https://docs.ccxt.com/en/latest/manual.html#fee-structure>` indexed by market symbols
        """
        self.load_markets()
        response = self.marketsGetSpotPairs(params)
        #
        #     {
        #         success: '1',
        #         data: {
        #           pairs: [
        #             {
        #               name: 'btc_jpy',
        #               base_asset: 'btc',
        #               quote_asset: 'jpy',
        #               maker_fee_rate_base: '0',
        #               taker_fee_rate_base: '0',
        #               maker_fee_rate_quote: '-0.0002',
        #               taker_fee_rate_quote: '0.0012',
        #               unit_amount: '0.0001',
        #               limit_max_amount: '1000',
        #               market_max_amount: '10',
        #               market_allowance_rate: '0.2',
        #               price_digits: '0',
        #               amount_digits: '4',
        #               is_enabled: True,
        #               stop_order: False,
        #               stop_order_and_cancel: False
        #             },
        #             ...
        #           ]
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        pairs = self.safe_value(data, 'pairs', [])
        result = {}
        for i in range(0, len(pairs)):
            pair = pairs[i]
            marketId = self.safe_string(pair, 'name')
            market = self.safe_market(marketId)
            symbol = market['symbol']
            result[symbol] = {
                'info': pair,
                'symbol': symbol,
                'maker': self.safe_number(pair, 'maker_fee_rate_quote'),
                'taker': self.safe_number(pair, 'taker_fee_rate_quote'),
                'percentage': True,
                'tierBased': False,
            }
        return result

    def parse_ohlcv(self, ohlcv, market=None):
        #
        #     [
        #         "0.02501786",
        #         "0.02501786",
        #         "0.02501786",
        #         "0.02501786",
        #         "0.0000",
        #         1591488000000
        #     ]
        #
        return [
            self.safe_integer(ohlcv, 5),
            self.safe_number(ohlcv, 0),
            self.safe_number(ohlcv, 1),
            self.safe_number(ohlcv, 2),
            self.safe_number(ohlcv, 3),
            self.safe_number(ohlcv, 4),
        ]

    def fetch_ohlcv(self, symbol, timeframe='1m', since=None, limit=None, params={}):
        """
        fetches historical candlestick data containing the open, high, low, and close price, and the volume of a market
        :param str symbol: unified symbol of the market to fetch OHLCV data for
        :param str timeframe: the length of time each candle represents
        :param int|None since: timestamp in ms of the earliest candle to fetch
        :param int|None limit: the maximum amount of candles to fetch
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns [[int]]: A list of candles ordered as timestamp, open, high, low, close, volume
        """
        if since is None:
            if limit is None:
                limit = 1000  # it doesn't have any defaults, might return 200, might 2000(i.e. https://public.bitbank.cc/btc_jpy/candlestick/4hour/2020)
            duration = self.parse_timeframe(timeframe)
            since = self.milliseconds() - duration * 1000 * limit
        self.load_markets()
        market = self.market(symbol)
        request = {
            'pair': market['id'],
            'candletype': self.timeframes[timeframe],
            'yyyymmdd': self.yyyymmdd(since, ''),
        }
        response = self.publicGetPairCandlestickCandletypeYyyymmdd(self.extend(request, params))
        #
        #     {
        #         "success":1,
        #         "data":{
        #             "candlestick":[
        #                 {
        #                     "type":"5min",
        #                     "ohlcv":[
        #                         ["0.02501786","0.02501786","0.02501786","0.02501786","0.0000",1591488000000],
        #                         ["0.02501747","0.02501953","0.02501747","0.02501953","0.3017",1591488300000],
        #                         ["0.02501762","0.02501762","0.02500392","0.02500392","0.1500",1591488600000],
        #                     ]
        #                 }
        #             ],
        #             "timestamp":1591508668190
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        candlestick = self.safe_value(data, 'candlestick', [])
        first = self.safe_value(candlestick, 0, {})
        ohlcv = self.safe_value(first, 'ohlcv', [])
        return self.parse_ohlcvs(ohlcv, market, timeframe, since, limit)

    def parse_balance(self, response):
        result = {
            'info': response,
            'timestamp': None,
            'datetime': None,
        }
        data = self.safe_value(response, 'data', {})
        assets = self.safe_value(data, 'assets', [])
        for i in range(0, len(assets)):
            balance = assets[i]
            currencyId = self.safe_string(balance, 'asset')
            code = self.safe_currency_code(currencyId)
            account = self.account()
            account['free'] = self.safe_string(balance, 'free_amount')
            account['used'] = self.safe_string(balance, 'locked_amount')
            account['total'] = self.safe_string(balance, 'onhand_amount')
            result[code] = account
        return self.safe_balance(result)

    def fetch_balance(self, params={}):
        """
        query for balance and get the amount of funds available for trading or funds locked in orders
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns dict: a `balance structure <https://docs.ccxt.com/en/latest/manual.html?#balance-structure>`
        """
        self.load_markets()
        response = self.privateGetUserAssets(params)
        #
        #     {
        #       "success": "1",
        #       "data": {
        #         "assets": [
        #           {
        #             "asset": "jpy",
        #             "amount_precision": "4",
        #             "onhand_amount": "0.0000",
        #             "locked_amount": "0.0000",
        #             "free_amount": "0.0000",
        #             "stop_deposit": False,
        #             "stop_withdrawal": False,
        #             "withdrawal_fee": {
        #               "threshold": "30000.0000",
        #               "under": "550.0000",
        #               "over": "770.0000"
        #             }
        #           },
        #           {
        #             "asset": "btc",
        #             "amount_precision": "8",
        #             "onhand_amount": "0.00000000",
        #             "locked_amount": "0.00000000",
        #             "free_amount": "0.00000000",
        #             "stop_deposit": False,
        #             "stop_withdrawal": False,
        #             "withdrawal_fee": "0.00060000"
        #           },
        #         ]
        #       }
        #     }
        #
        return self.parse_balance(response)

    def parse_order_status(self, status):
        statuses = {
            'UNFILLED': 'open',
            'PARTIALLY_FILLED': 'open',
            'FULLY_FILLED': 'closed',
            'CANCELED_UNFILLED': 'canceled',
            'CANCELED_PARTIALLY_FILLED': 'canceled',
        }
        return self.safe_string(statuses, status, status)

    def parse_order(self, order, market=None):
        id = self.safe_string(order, 'order_id')
        marketId = self.safe_string(order, 'pair')
        market = self.safe_market(marketId, market)
        timestamp = self.safe_integer(order, 'ordered_at')
        price = self.safe_string(order, 'price')
        amount = self.safe_string(order, 'start_amount')
        filled = self.safe_string(order, 'executed_amount')
        remaining = self.safe_string(order, 'remaining_amount')
        average = self.safe_string(order, 'average_price')
        status = self.parse_order_status(self.safe_string(order, 'status'))
        type = self.safe_string_lower(order, 'type')
        side = self.safe_string_lower(order, 'side')
        return self.safe_order({
            'id': id,
            'clientOrderId': None,
            'datetime': self.iso8601(timestamp),
            'timestamp': timestamp,
            'lastTradeTimestamp': None,
            'status': status,
            'symbol': market['symbol'],
            'type': type,
            'timeInForce': None,
            'postOnly': None,
            'side': side,
            'price': price,
            'stopPrice': None,
            'cost': None,
            'average': average,
            'amount': amount,
            'filled': filled,
            'remaining': remaining,
            'trades': None,
            'fee': None,
            'info': order,
        }, market)

    def create_order(self, symbol, type, side, amount, price=None, params={}):
        """
        create a trade order
        :param str symbol: unified symbol of the market to create an order in
        :param str type: 'market' or 'limit'
        :param str side: 'buy' or 'sell'
        :param float amount: how much of currency you want to trade in units of base currency
        :param float|None price: the price at which the order is to be fullfilled, in units of the quote currency, ignored in market orders
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns dict: an `order structure <https://docs.ccxt.com/en/latest/manual.html#order-structure>`
        """
        self.load_markets()
        market = self.market(symbol)
        request = {
            'pair': market['id'],
            'amount': self.amount_to_precision(symbol, amount),
            'side': side,
            'type': type,
        }
        if type == 'limit':
            request['price'] = self.price_to_precision(symbol, price)
        response = self.privatePostUserSpotOrder(self.extend(request, params))
        data = self.safe_value(response, 'data')
        return self.parse_order(data, market)

    def cancel_order(self, id, symbol=None, params={}):
        """
        cancels an open order
        :param str id: order id
        :param str|None symbol: unified symbol of the market the order was made in
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns dict: An `order structure <https://docs.ccxt.com/en/latest/manual.html#order-structure>`
        """
        self.load_markets()
        market = self.market(symbol)
        request = {
            'order_id': id,
            'pair': market['id'],
        }
        response = self.privatePostUserSpotCancelOrder(self.extend(request, params))
        data = self.safe_value(response, 'data')
        return data

    def fetch_order(self, id, symbol=None, params={}):
        """
        fetches information on an order made by the user
        :param str|None symbol: unified symbol of the market the order was made in
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns dict: An `order structure <https://docs.ccxt.com/en/latest/manual.html#order-structure>`
        """
        self.load_markets()
        market = self.market(symbol)
        request = {
            'order_id': id,
            'pair': market['id'],
        }
        response = self.privateGetUserSpotOrder(self.extend(request, params))
        data = self.safe_value(response, 'data')
        return self.parse_order(data, market)

    def fetch_open_orders(self, symbol=None, since=None, limit=None, params={}):
        """
        fetch all unfilled currently open orders
        :param str|None symbol: unified market symbol
        :param int|None since: the earliest time in ms to fetch open orders for
        :param int|None limit: the maximum number of  open orders structures to retrieve
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns [dict]: a list of `order structures <https://docs.ccxt.com/en/latest/manual.html#order-structure>`
        """
        self.load_markets()
        market = self.market(symbol)
        request = {
            'pair': market['id'],
        }
        if limit is not None:
            request['count'] = limit
        if since is not None:
            request['since'] = int(since / 1000)
        response = self.privateGetUserSpotActiveOrders(self.extend(request, params))
        data = self.safe_value(response, 'data', {})
        orders = self.safe_value(data, 'orders', [])
        return self.parse_orders(orders, market, since, limit)

    def fetch_my_trades(self, symbol=None, since=None, limit=None, params={}):
        """
        fetch all trades made by the user
        :param str|None symbol: unified market symbol
        :param int|None since: the earliest time in ms to fetch trades for
        :param int|None limit: the maximum number of trades structures to retrieve
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns [dict]: a list of `trade structures <https://docs.ccxt.com/en/latest/manual.html#trade-structure>`
        """
        self.load_markets()
        request = {}
        market = None
        if symbol is not None:
            market = self.market(symbol)
            request['pair'] = market['id']
        if limit is not None:
            request['count'] = limit
        if since is not None:
            request['since'] = int(since / 1000)
        response = self.privateGetUserSpotTradeHistory(self.extend(request, params))
        data = self.safe_value(response, 'data', {})
        trades = self.safe_value(data, 'trades', [])
        return self.parse_trades(trades, market, since, limit)

    def fetch_deposit_address(self, code, params={}):
        """
        fetch the deposit address for a currency associated with self account
        :param str code: unified currency code
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns dict: an `address structure <https://docs.ccxt.com/en/latest/manual.html#address-structure>`
        """
        self.load_markets()
        currency = self.currency(code)
        request = {
            'asset': currency['id'],
        }
        response = self.privateGetUserWithdrawalAccount(self.extend(request, params))
        data = self.safe_value(response, 'data', {})
        # Not sure about self if there could be more than one account...
        accounts = self.safe_value(data, 'accounts', [])
        firstAccount = self.safe_value(accounts, 0, {})
        address = self.safe_string(firstAccount, 'address')
        return {
            'currency': currency,
            'address': address,
            'tag': None,
            'network': None,
            'info': response,
        }

    def withdraw(self, code, amount, address, tag=None, params={}):
        """
        make a withdrawal
        :param str code: unified currency code
        :param float amount: the amount to withdraw
        :param str address: the address to withdraw to
        :param str|None tag:
        :param dict params: extra parameters specific to the bitbank api endpoint
        :returns dict: a `transaction structure <https://docs.ccxt.com/en/latest/manual.html#transaction-structure>`
        """
        tag, params = self.handle_withdraw_tag_and_params(tag, params)
        if not ('uuid' in params):
            raise ExchangeError(self.id + ' uuid is required for withdrawal')
        self.load_markets()
        currency = self.currency(code)
        request = {
            'asset': currency['id'],
            'amount': amount,
        }
        response = self.privatePostUserRequestWithdrawal(self.extend(request, params))
        #
        #     {
        #         "success": 1,
        #         "data": {
        #             "uuid": "string",
        #             "asset": "btc",
        #             "amount": 0,
        #             "account_uuid": "string",
        #             "fee": 0,
        #             "status": "DONE",
        #             "label": "string",
        #             "txid": "string",
        #             "address": "string",
        #             "requested_at": 0
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        return self.parse_transaction(data, currency)

    def parse_transaction(self, transaction, currency=None):
        #
        # withdraw
        #
        #     {
        #         "uuid": "string",
        #         "asset": "btc",
        #         "amount": 0,
        #         "account_uuid": "string",
        #         "fee": 0,
        #         "status": "DONE",
        #         "label": "string",
        #         "txid": "string",
        #         "address": "string",
        #         "requested_at": 0
        #     }
        #
        txid = self.safe_string(transaction, 'txid')
        currency = self.safe_currency(None, currency)
        return {
            'id': txid,
            'txid': txid,
            'timestamp': None,
            'datetime': None,
            'network': None,
            'addressFrom': None,
            'address': None,
            'addressTo': None,
            'amount': None,
            'type': None,
            'currency': currency['code'],
            'status': None,
            'updated': None,
            'tagFrom': None,
            'tag': None,
            'tagTo': None,
            'comment': None,
            'fee': None,
            'info': transaction,
        }

    def nonce(self):
        return self.milliseconds()

    def sign(self, path, api='public', method='GET', params={}, headers=None, body=None):
        query = self.omit(params, self.extract_params(path))
        url = self.implode_hostname(self.urls['api'][api]) + '/'
        if (api == 'public') or (api == 'markets'):
            url += self.implode_params(path, params)
            if query:
                url += '?' + self.urlencode(query)
        else:
            self.check_required_credentials()
            nonce = str(self.nonce())
            auth = nonce
            url += self.version + '/' + self.implode_params(path, params)
            if method == 'POST':
                body = self.json(query)
                auth += body
            else:
                auth += '/' + self.version + '/' + path
                if query:
                    query = self.urlencode(query)
                    url += '?' + query
                    auth += '?' + query
            headers = {
                'Content-Type': 'application/json',
                'ACCESS-KEY': self.apiKey,
                'ACCESS-NONCE': nonce,
                'ACCESS-SIGNATURE': self.hmac(self.encode(auth), self.encode(self.secret)),
            }
        return {'url': url, 'method': method, 'body': body, 'headers': headers}

    def handle_errors(self, httpCode, reason, url, method, headers, body, response, requestHeaders, requestBody):
        if response is None:
            return
        success = self.safe_integer(response, 'success')
        data = self.safe_value(response, 'data')
        if not success or not data:
            errorMessages = {
                '10000': 'URL does not exist',
                '10001': 'A system error occurred. Please contact support',
                '10002': 'Invalid JSON format. Please check the contents of transmission',
                '10003': 'A system error occurred. Please contact support',
                '10005': 'A timeout error occurred. Please wait for a while and try again',
                '20001': 'API authentication failed',
                '20002': 'Illegal API key',
                '20003': 'API key does not exist',
                '20004': 'API Nonce does not exist',
                '20005': 'API signature does not exist',
                '20011': 'Two-step verification failed',
                '20014': 'SMS authentication failed',
                '30001': 'Please specify the order quantity',
                '30006': 'Please specify the order ID',
                '30007': 'Please specify the order ID array',
                '30009': 'Please specify the stock',
                '30012': 'Please specify the order price',
                '30013': 'Trade Please specify either',
                '30015': 'Please specify the order type',
                '30016': 'Please specify asset name',
                '30019': 'Please specify uuid',
                '30039': 'Please specify the amount to be withdrawn',
                '40001': 'The order quantity is invalid',
                '40006': 'Count value is invalid',
                '40007': 'End time is invalid',
                '40008': 'end_id Value is invalid',
                '40009': 'The from_id value is invalid',
                '40013': 'The order ID is invalid',
                '40014': 'The order ID array is invalid',
                '40015': 'Too many specified orders',
                '40017': 'Incorrect issue name',
                '40020': 'The order price is invalid',
                '40021': 'The trading classification is invalid',
                '40022': 'Start date is invalid',
                '40024': 'The order type is invalid',
                '40025': 'Incorrect asset name',
                '40028': 'uuid is invalid',
                '40048': 'The amount of withdrawal is illegal',
                '50003': 'Currently, self account is in a state where you can not perform the operation you specified. Please contact support',
                '50004': 'Currently, self account is temporarily registered. Please try again after registering your account',
                '50005': 'Currently, self account is locked. Please contact support',
                '50006': 'Currently, self account is locked. Please contact support',
                '50008': 'User identification has not been completed',
                '50009': 'Your order does not exist',
                '50010': 'Can not cancel specified order',
                '50011': 'API not found',
                '60001': 'The number of possessions is insufficient',
                '60002': 'It exceeds the quantity upper limit of the tender buying order',
                '60003': 'The specified quantity exceeds the limit',
                '60004': 'The specified quantity is below the threshold',
                '60005': 'The specified price is above the limit',
                '60006': 'The specified price is below the lower limit',
                '70001': 'A system error occurred. Please contact support',
                '70002': 'A system error occurred. Please contact support',
                '70003': 'A system error occurred. Please contact support',
                '70004': 'We are unable to accept orders as the transaction is currently suspended',
                '70005': 'Order can not be accepted because purchase order is currently suspended',
                '70006': 'We can not accept orders because we are currently unsubscribed ',
                '70009': 'We are currently temporarily restricting orders to be carried out. Please use the limit order.',
                '70010': 'We are temporarily raising the minimum order quantity as the system load is now rising.',
            }
            errorClasses = self.exceptions
            code = self.safe_string(data, 'code')
            message = self.safe_string(errorMessages, code, 'Error')
            ErrorClass = self.safe_value(errorClasses, code)
            if ErrorClass is not None:
                raise ErrorClass(message)
            else:
                raise ExchangeError(self.id + ' ' + self.json(response))



# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

from ccxt.pro.base.exchange import Exchange
import ccxt.async_support
from ccxt.pro.base.cache import ArrayCache
from ccxt.base.errors import ExchangeError


class zb(Exchange, ccxt.async_support.zb):

    def describe(self):
        return self.deep_extend(super(zb, self).describe(), {
            'has': {
                'ws': True,
                'watchOrderBook': True,
                'watchTicker': True,
                'watchTrades': True,
            },
            'urls': {
                'api': {
                    'ws': 'wss://api.{hostname}/websocket',
                },
            },
            'options': {
                'tradesLimit': 1000,
                'ordersLimit': 1000,
                'OHLCVLimit': 1000,
            },
        })

    async def watch_public(self, name, symbol, method, params={}):
        await self.load_markets()
        market = self.market(symbol)
        symbol = market['symbol']
        messageHash = market['baseId'] + market['quoteId'] + '_' + name
        url = self.implode_hostname(self.urls['api']['ws'])
        request = {
            'event': 'addChannel',
            'channel': messageHash,
        }
        message = self.extend(request, params)
        subscription = {
            'name': name,
            'symbol': symbol,
            'marketId': market['id'],
            'messageHash': messageHash,
            'method': method,
        }
        return await self.watch(url, messageHash, message, messageHash, subscription)

    async def watch_ticker(self, symbol, params={}):
        """
        watches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific market
        :param str symbol: unified symbol of the market to fetch the ticker for
        :param dict params: extra parameters specific to the zb api endpoint
        :returns dict: a `ticker structure <https://docs.ccxt.com/en/latest/manual.html#ticker-structure>`
        """
        return await self.watch_public('ticker', symbol, self.handle_ticker, params)

    def handle_ticker(self, client, message, subscription):
        #
        #     {
        #         date: '1624398991255',
        #         ticker: {
        #             high: '33298.38',
        #             vol: '56375.9469',
        #             last: '32396.95',
        #             low: '28808.19',
        #             buy: '32395.81',
        #             sell: '32409.3',
        #             turnover: '1771122527.0000',
        #             open: '31652.44',
        #             riseRate: '2.36'
        #         },
        #         dataType: 'ticker',
        #         channel: 'btcusdt_ticker'
        #     }
        #
        symbol = self.safe_string(subscription, 'symbol')
        channel = self.safe_string(message, 'channel')
        market = self.market(symbol)
        data = self.safe_value(message, 'ticker')
        data['date'] = self.safe_value(message, 'date')
        ticker = self.parse_ticker(data, market)
        ticker['symbol'] = symbol
        self.tickers[symbol] = ticker
        client.resolve(ticker, channel)
        return message

    async def watch_trades(self, symbol, since=None, limit=None, params={}):
        """
        get the list of most recent trades for a particular symbol
        :param str symbol: unified symbol of the market to fetch trades for
        :param int|None since: timestamp in ms of the earliest trade to fetch
        :param int|None limit: the maximum amount of trades to fetch
        :param dict params: extra parameters specific to the zb api endpoint
        :returns [dict]: a list of `trade structures <https://docs.ccxt.com/en/latest/manual.html?#public-trades>`
        """
        await self.load_markets()
        symbol = self.symbol(symbol)
        trades = await self.watch_public('trades', symbol, self.handle_trades, params)
        if self.newUpdates:
            limit = trades.getLimit(symbol, limit)
        return self.filter_by_since_limit(trades, since, limit, 'timestamp', True)

    def handle_trades(self, client, message, subscription):
        #
        #     {
        #         data: [
        #             {date: 1624537147, amount: '0.0357', price: '34066.11', trade_type: 'bid', type: 'buy', tid: 1718857158},
        #             {date: 1624537147, amount: '0.0255', price: '34071.04', trade_type: 'bid', type: 'buy', tid: 1718857159},
        #             {date: 1624537147, amount: '0.0153', price: '34071.29', trade_type: 'bid', type: 'buy', tid: 1718857160}
        #         ],
        #         dataType: 'trades',
        #         channel: 'btcusdt_trades'
        #     }
        #
        channel = self.safe_value(message, 'channel')
        symbol = self.safe_string(subscription, 'symbol')
        market = self.market(symbol)
        data = self.safe_value(message, 'data')
        trades = self.parse_trades(data, market)
        tradesArray = self.safe_value(self.trades, symbol)
        if tradesArray is None:
            limit = self.safe_integer(self.options, 'tradesLimit', 1000)
            tradesArray = ArrayCache(limit)
        for i in range(0, len(trades)):
            tradesArray.append(trades[i])
        self.trades[symbol] = tradesArray
        client.resolve(tradesArray, channel)

    async def watch_order_book(self, symbol, limit=None, params={}):
        """
        watches information on open orders with bid(buy) and ask(sell) prices, volumes and other data
        :param str symbol: unified symbol of the market to fetch the order book for
        :param int|None limit: the maximum amount of order book entries to return
        :param dict params: extra parameters specific to the zb api endpoint
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/en/latest/manual.html#order-book-structure>` indexed by market symbols
        """
        if limit is not None:
            if (limit != 5) and (limit != 10) and (limit != 20):
                raise ExchangeError(self.id + ' watchOrderBook limit argument must be None, 5, 10 or 20')
        else:
            limit = 5  # default
        await self.load_markets()
        market = self.market(symbol)
        symbol = market['symbol']
        name = 'quick_depth'
        messageHash = market['baseId'] + market['quoteId'] + '_' + name
        url = self.implode_hostname(self.urls['api']['ws']) + '/' + market['baseId']
        request = {
            'event': 'addChannel',
            'channel': messageHash,
            'length': limit,
        }
        message = self.extend(request, params)
        subscription = {
            'name': name,
            'symbol': symbol,
            'marketId': market['id'],
            'messageHash': messageHash,
            'method': self.handle_order_book,
        }
        orderbook = await self.watch(url, messageHash, message, messageHash, subscription)
        return orderbook.limit()

    def handle_order_book(self, client, message, subscription):
        #
        #     {
        #         lastTime: 1624524640066,
        #         dataType: 'quickDepth',
        #         channel: 'btcusdt_quick_depth',
        #         currentPrice: 33183.79,
        #         listDown: [
        #             [33166.87, 0.2331],
        #             [33166.86, 0.15],
        #             [33166.76, 0.15],
        #             [33161.02, 0.212],
        #             [33146.35, 0.6066]
        #         ],
        #         market: 'btcusdt',
        #         listUp: [
        #             [33186.88, 0.15],
        #             [33190.1, 0.15],
        #             [33193.03, 0.2518],
        #             [33195.05, 0.2031],
        #             [33199.99, 0.6066]
        #         ],
        #         high: 34816.8,
        #         rate: '6.484',
        #         low: 32312.41,
        #         currentIsBuy: True,
        #         dayNumber: 26988.5536,
        #         totalBtc: 26988.5536,
        #         showMarket: 'btcusdt'
        #     }
        #
        channel = self.safe_string(message, 'channel')
        limit = self.safe_integer(subscription, 'limit')
        symbol = self.safe_string(subscription, 'symbol')
        orderbook = self.safe_value(self.orderbooks, symbol)
        if orderbook is None:
            orderbook = self.order_book({}, limit)
            self.orderbooks[symbol] = orderbook
        timestamp = self.safe_integer(message, 'lastTime')
        parsed = self.parse_order_book(message, symbol, timestamp, 'listDown', 'listUp')
        orderbook.reset(parsed)
        orderbook['symbol'] = symbol
        client.resolve(orderbook, channel)

    def handle_message(self, client, message):
        #
        #
        #     {
        #         no: '0',
        #         code: 1007,
        #         success: False,
        #         channel: 'btc_usdt_ticker',
        #         message: 'Channel is empty'
        #     }
        #
        #     {
        #         date: '1624398991255',
        #         ticker: {
        #             high: '33298.38',
        #             vol: '56375.9469',
        #             last: '32396.95',
        #             low: '28808.19',
        #             buy: '32395.81',
        #             sell: '32409.3',
        #             turnover: '1771122527.0000',
        #             open: '31652.44',
        #             riseRate: '2.36'
        #         },
        #         dataType: 'ticker',
        #         channel: 'btcusdt_ticker'
        #     }
        #
        #     {
        #         data: [
        #             {date: 1624537147, amount: '0.0357', price: '34066.11', trade_type: 'bid', type: 'buy', tid: 1718857158},
        #             {date: 1624537147, amount: '0.0255', price: '34071.04', trade_type: 'bid', type: 'buy', tid: 1718857159},
        #             {date: 1624537147, amount: '0.0153', price: '34071.29', trade_type: 'bid', type: 'buy', tid: 1718857160}
        #         ],
        #         dataType: 'trades',
        #         channel: 'btcusdt_trades'
        #     }
        #
        dataType = self.safe_string(message, 'dataType')
        if dataType is not None:
            channel = self.safe_string(message, 'channel')
            subscription = self.safe_value(client.subscriptions, channel)
            if subscription is not None:
                method = self.safe_value(subscription, 'method')
                if method is not None:
                    return method(client, message, subscription)
            return message



# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------

__version__ = '2.1.33'

# -----------------------------------------------------------------------------

from ccxt.pro.base.functions import inflate, inflate64, gunzip
from ccxt.pro.base.fast_client import FastClient
from ccxt.async_support.base.exchange import Exchange as BaseExchange
from ccxt import NotSupported
from ccxt.pro.base.order_book import OrderBook, IndexedOrderBook, CountedOrderBook
from ccxt.async_support.base.throttler import Throttler
import asyncio

# -----------------------------------------------------------------------------

__all__ = [
    'BaseExchange',
    'Exchange',
]

# -----------------------------------------------------------------------------


class Exchange(BaseExchange):

    clients = {}

    # streaming-specific options
    streaming = {
        'keepAlive': 30000,
        'ping': None,
        'maxPingPongMisses': 2.0,
    }

    newUpdates = True

    @staticmethod
    def inflate(data):
        return inflate(data)

    @staticmethod
    def inflate64(data):
        return inflate64(data)

    @staticmethod
    def gunzip(data):
        return gunzip(data)

    def order_book(self, snapshot={}, depth=None):
        return OrderBook(snapshot, depth)

    def indexed_order_book(self, snapshot={}, depth=None):
        return IndexedOrderBook(snapshot, depth)

    def counted_order_book(self, snapshot={}, depth=None):
        return CountedOrderBook(snapshot, depth)

    def client(self, url):
        self.clients = self.clients or {}
        if url not in self.clients:
            on_message = self.handle_message
            on_error = self.on_error
            on_close = self.on_close
            on_connected = self.on_connected
            # decide client type here: aiohttp ws / websockets / signalr / socketio
            ws_options = self.safe_value(self.options, 'ws', {})
            options = self.extend(self.streaming, {
                'log': getattr(self, 'log'),
                'ping': getattr(self, 'ping', None),
                'verbose': self.verbose,
                'throttle': Throttler(self.tokenBucket, self.asyncio_loop),
                'asyncio_loop': self.asyncio_loop,
            }, ws_options)
            self.clients[url] = FastClient(url, on_message, on_error, on_close, on_connected, options)
        return self.clients[url]

    async def spawn_async(self, method, *args):
        try:
            await method(*args)
        except Exception:
            # todo: handle spawned errors
            pass

    async def delay_async(self, timeout, method, *args):
        await self.sleep(timeout)
        try:
            await method(*args)
        except Exception:
            # todo: handle spawned errors
            pass

    def spawn(self, method, *args):
        asyncio.ensure_future(self.spawn_async(method, *args))

    def delay(self, timeout, method, *args):
        asyncio.ensure_future(self.delay_async(timeout, method, *args))

    def handle_message(self, client, message):
        always = True
        if always:
            raise NotSupported(self.id + '.handle_message() not implemented yet')
        return {}

    def watch(self, url, message_hash, message=None, subscribe_hash=None, subscription=None):
        backoff_delay = 0
        client = self.client(url)
        future = client.future(message_hash)

        # base exchange self.open starts the aiohttp Session in an async context
        self.open()
        connected = client.connected if client.connected.done() \
            else asyncio.ensure_future(client.connect(self.session, backoff_delay))

        def after(fut):
            if subscribe_hash not in client.subscriptions:
                client.subscriptions[subscribe_hash] = subscription or True
                # todo: decouple signing from subscriptions
                options = self.safe_value(self.options, 'ws')
                cost = self.safe_value(options, 'cost', 1)
                if message:
                    async def send_message():
                        if self.enableRateLimit:
                            await client.throttle(cost)
                        try:
                            await client.send(message)
                        except ConnectionError as e:
                            future.reject(e)
                    asyncio.ensure_future(send_message())

        connected.add_done_callback(after)

        return future

    def on_connected(self, client, message=None):
        # for user hooks
        # print('Connected to', client.url)
        pass

    def on_error(self, client, error):
        if client.url in self.clients and self.clients[client.url].error:
            del self.clients[client.url]

    def on_close(self, client, error):
        if client.error:
            # connection closed due to an error
            pass
        else:
            # server disconnected a working connection
            if client.url in self.clients:
                del self.clients[client.url]

    async def close(self):
        if self.clients:
            await asyncio.wait([asyncio.create_task(client.close()) for client in self.clients.values()], return_when=asyncio.ALL_COMPLETED)
            for url in self.clients.copy():
                del self.clients[url]
        await super(Exchange, self).close()

    def find_timeframe(self, timeframe, timeframes=None):
        timeframes = timeframes if timeframes else self.timeframes
        for key, value in timeframes.items():
            if value == timeframe:
                return key
        return None

    def format_scientific_notation_ftx(self, n):
        if n == 0:
            return '0e-00'
        return format(n, 'g')

    async def watch_ticker(self, symbol, params={}):
        raise NotSupported(self.id + '.watch_ticker() not implemented yet')

    async def watch_order_book(self, symbol, limit=None, params={}):
        raise NotSupported(self.id + '.watch_order_book() not implemented yet')

    async def watch_trades(self, symbol, since=None, limit=None, params={}):
        raise NotSupported(self.id + '.watch_trades() not implemented yet')

    async def watch_ohlcv(self, symbol, timeframe='1m', since=None, limit=None, params={}):
        raise NotSupported(self.id + '.watch_ohlcv() not implemented yet')

    async def watch_balance(self, params={}):
        raise NotSupported(self.id + '.watch_balance() not implemented yet')

    async def watch_orders(self, symbol=None, since=None, limit=None, params={}):
        raise NotSupported(self.id + '.watch_orders() not implemented yet')

    async def watch_my_trades(self, symbol=None, since=None, limit=None, params={}):
        raise NotSupported(self.id + '.watch_my_trades() not implemented yet')



# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

from ccxt.base.exchange import Exchange
from ccxt.base.errors import ExchangeError
from ccxt.base.errors import AuthenticationError
from ccxt.base.errors import BadSymbol
from ccxt.base.errors import InsufficientFunds
from ccxt.base.errors import InvalidOrder
from ccxt.base.errors import RateLimitExceeded
from ccxt.base.decimal_to_precision import TICK_SIZE
from ccxt.base.precise import Precise


class qtrade(Exchange):

    def describe(self):
        return self.deep_extend(super(qtrade, self).describe(), {
            'id': 'qtrade',
            'name': 'qTrade',
            'countries': ['US'],
            'rateLimit': 1000,
            'version': 'v1',
            'urls': {
                'logo': 'https://user-images.githubusercontent.com/51840849/80491487-74a99c00-896b-11ea-821e-d307e832f13e.jpg',
                'api': {
                    'rest': 'https://api.qtrade.io',
                },
                'www': 'https://qtrade.io',
                'doc': 'https://qtrade-exchange.github.io/qtrade-docs',
                'referral': 'https://qtrade.io/?ref=BKOQWVFGRH2C',
                'fees': 'https://qtrade.io/fees',
            },
            'has': {
                'CORS': None,
                'spot': True,
                'margin': False,
                'swap': False,
                'future': False,
                'option': False,
                'addMargin': False,
                'cancelOrder': True,
                'createMarketOrder': None,
                'createOrder': True,
                'createReduceOnlyOrder': False,
                'createStopLimitOrder': False,
                'createStopMarketOrder': False,
                'createStopOrder': False,
                'fetchBalance': True,
                'fetchBorrowRate': False,
                'fetchBorrowRateHistories': False,
                'fetchBorrowRateHistory': False,
                'fetchBorrowRates': False,
                'fetchBorrowRatesPerSymbol': False,
                'fetchClosedOrders': True,
                'fetchCurrencies': True,
                'fetchDeposit': True,
                'fetchDepositAddress': True,
                'fetchDeposits': True,
                'fetchFundingHistory': False,
                'fetchFundingRate': False,
                'fetchFundingRateHistory': False,
                'fetchFundingRates': False,
                'fetchIndexOHLCV': False,
                'fetchLeverage': False,
                'fetchLeverageTiers': False,
                'fetchMarginMode': False,
                'fetchMarkets': True,
                'fetchMarkOHLCV': False,
                'fetchMyTrades': True,
                'fetchOHLCV': True,
                'fetchOpenInterestHistory': False,
                'fetchOpenOrders': True,
                'fetchOrder': True,
                'fetchOrderBook': True,
                'fetchOrders': True,
                'fetchPosition': False,
                'fetchPositionMode': False,
                'fetchPositions': False,
                'fetchPositionsRisk': False,
                'fetchPremiumIndexOHLCV': False,
                'fetchTicker': True,
                'fetchTickers': True,
                'fetchTrades': True,
                'fetchTradingFee': True,
                'fetchTradingFees': False,
                'fetchTransactions': None,
                'fetchTransfer': False,
                'fetchTransfers': True,
                'fetchWithdrawal': True,
                'fetchWithdrawals': True,
                'reduceMargin': False,
                'setLeverage': False,
                'setMarginMode': False,
                'setPositionMode': False,
                'transfer': False,
                'withdraw': True,
            },
            'timeframes': {
                '5m': 'fivemin',
                '15m': 'fifteenmin',
                '30m': 'thirtymin',
                '1h': 'onehour',
                '2h': 'twohour',
                '4h': 'fourhour',
                '1d': 'oneday',
            },
            'api': {
                'public': {
                    'get': [
                        'ticker/{market_string}',
                        'tickers',
                        'currency/{code}',
                        'currencies',
                        'common',
                        'market/{market_string}',
                        'markets',
                        'market/{market_string}/trades',
                        'orderbook/{market_string}',
                        'market/{market_string}/ohlcv/{interval}',
                    ],
                },
                'private': {
                    'get': [
                        'me',
                        'balances',
                        'balances_all',  # undocumented
                        'market/{market_string}',
                        'orders',
                        'order/{order_id}',
                        'trades',
                        'withdraw/{withdraw_id}',
                        'withdraws',
                        'deposit/{deposit_id}',
                        'deposits',
                        'transfers',
                    ],
                    'post': [
                        'cancel_order',
                        'withdraw',
                        'deposit_address/{currency}',
                        'sell_limit',
                        'buy_limit',
                    ],
                },
            },
            'fees': {
                'trading': {
                    'feeSide': 'quote',
                    'tierBased': True,
                    'percentage': True,
                    'taker': self.parse_number('0.005'),
                    'maker': self.parse_number('0.0'),
                },
                'funding': {
                    'withdraw': {},
                },
            },
            'commonCurrencies': {
                'BTM': 'Bitmark',
            },
            'precisionMode': TICK_SIZE,
            'exceptions': {
                'exact': {
                    'invalid_auth': AuthenticationError,
                    'insuff_funds': InsufficientFunds,
                    'market_not_found': BadSymbol,  # {"errors":[{"code":"market_not_found","title":"Requested market does not exist"}]}
                    'too_small': InvalidOrder,
                    'limit_exceeded': RateLimitExceeded,  # {"errors":[{"code":"limit_exceeded","title":"You have exceeded the windowed rate limit. Please see docs."}]}
                },
            },
        })

    def fetch_markets(self, params={}):
        """
        retrieves data on all markets for qtrade
        :param dict params: extra parameters specific to the exchange api endpoint
        :returns [dict]: an array of objects representing market data
        """
        response = self.publicGetMarkets(params)
        #
        #     {
        #         "data":{
        #             "markets":[
        #                 {
        #                     "id":5,
        #                     "market_currency":"BAC",
        #                     "base_currency":"BTC",
        #                     "maker_fee":"0.0025",
        #                     "taker_fee":"0.0025",
        #                     "metadata":{
        #                         "delisting_date":"7/15/2018",
        #                         "market_notices":[
        #                             {
        #                                 "message":"Delisting Notice: This market has been delisted due to low volume. Please cancel your orders and withdraw your funds by 7/15/2018.",
        #                                 "type":"warning"
        #                             }
        #                         ]
        #                     },
        #                     "can_trade":false,
        #                     "can_cancel":true,
        #                     "can_view":false,
        #                     "market_string":"BAC_BTC",
        #                     "minimum_sell_amount":"0.0001",
        #                     "minimum_buy_value":"0.0001",
        #                     "market_precision":8,  # note, they have reversed understanding of 'quote' vs 'base' concepts
        #                     "base_precision":8  # as noted in above comment
        #                 },
        #             ],
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        markets = self.safe_value(data, 'markets', [])
        result = []
        for i in range(0, len(markets)):
            market = markets[i]
            marketId = self.safe_string(market, 'market_string')
            numericId = self.safe_integer(market, 'id')
            baseId = self.safe_string(market, 'market_currency')
            quoteId = self.safe_string(market, 'base_currency')
            base = self.safe_currency_code(baseId)
            quote = self.safe_currency_code(quoteId)
            canView = self.safe_value(market, 'can_view', False)
            canTrade = self.safe_value(market, 'can_trade', False)
            active = canTrade and canView
            result.append({
                'id': marketId,
                'numericId': numericId,
                'symbol': base + '/' + quote,
                'base': base,
                'quote': quote,
                'settle': None,
                'baseId': baseId,
                'quoteId': quoteId,
                'settleId': None,
                'type': 'spot',
                'spot': True,
                'margin': False,
                'swap': False,
                'future': False,
                'option': False,
                'active': active,
                'contract': False,
                'linear': None,
                'inverse': None,
                'taker': self.safe_number(market, 'taker_fee'),
                'maker': self.safe_number(market, 'maker_fee'),
                'contractSize': None,
                'expiry': None,
                'expiryDatetime': None,
                'strike': None,
                'optionType': None,
                'precision': {
                    'amount': self.parse_number(self.parse_precision(self.safe_string(market, 'market_precision'))),
                    'price': self.parse_number(self.parse_precision(self.safe_string(market, 'base_precision'))),
                },
                'limits': {
                    'leverage': {
                        'min': None,
                        'max': None,
                    },
                    'amount': {
                        'min': self.safe_number(market, 'minimum_sell_value'),
                        'max': None,
                    },
                    'price': {
                        'min': None,
                        'max': None,
                    },
                    'cost': {
                        'min': self.safe_number(market, 'minimum_buy_value'),
                        'max': None,
                    },
                },
                'info': market,
            })
        return result

    def fetch_currencies(self, params={}):
        """
        fetches all available currencies on an exchange
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: an associative dictionary of currencies
        """
        response = self.publicGetCurrencies(params)
        #
        #     {
        #         "data":{
        #             "currencies":[
        #                 {
        #                     "code":"DGB",
        #                     "long_name":"Digibyte",
        #                     "type":"bitcoin_like",
        #                     "precision":8,
        #                     "config":{
        #                         "price":0.0035,
        #                         "withdraw_fee":"10",
        #                         "deposit_types":[
        #                             {
        #                                 "label":"Address",
        #                                 "lookup_mode":"address",
        #                                 "render_type":"address",
        #                                 "deposit_type":"address",
        #                                 "lookup_config":{}
        #                             }
        #                         ],
        #                         "default_signer":103,
        #                         "address_version":30,
        #                         "satoshi_per_byte":300,
        #                         "required_confirmations":200,
        #                         "required_generate_confirmations":300
        #                     },
        #                     "metadata":{},
        #                     "minimum_order":"0.0001",
        #                     "status":"ok",
        #                     "can_withdraw":true,
        #                     "delisted":false,
        #                     "deposit_disabled":false,
        #                     "withdraw_disabled":false,
        #                     "deposit_warn_codes":[],
        #                     "withdraw_warn_codes":[]
        #                 },
        #             ],
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        currencies = self.safe_value(data, 'currencies', [])
        result = {}
        for i in range(0, len(currencies)):
            currency = currencies[i]
            id = self.safe_string(currency, 'code')
            code = self.safe_currency_code(id)
            name = self.safe_string(currency, 'long_name')
            type = self.safe_string(currency, 'type')
            canWithdraw = self.safe_value(currency, 'can_withdraw', True)
            withdrawDisabled = self.safe_value(currency, 'withdraw_disabled', False)
            depositDisabled = self.safe_value(currency, 'deposit_disabled', False)
            deposit = not depositDisabled
            withdraw = canWithdraw and not withdrawDisabled
            config = self.safe_value(currency, 'config', {})
            status = self.safe_string(currency, 'status')
            active = withdraw and deposit and (status == 'ok')
            result[code] = {
                'id': id,
                'code': code,
                'info': currency,
                'type': type,
                'name': name,
                'fee': self.safe_number(config, 'withdraw_fee'),
                'precision': self.parse_number(self.parse_precision(self.safe_string(currency, 'precision'))),
                'active': active,
                'deposit': deposit,
                'withdraw': withdraw,
                'limits': {
                    'amount': {
                        'min': self.safe_number(currency, 'minimum_order'),
                        'max': None,
                    },
                    'withdraw': {
                        'min': None,
                        'max': None,
                    },
                },
            }
        return result

    def parse_ohlcv(self, ohlcv, market=None):
        #
        #     {
        #         "time":"2019-12-07T22:55:00Z",
        #         "open":"0.00197",
        #         "high":"0.00197",
        #         "low":"0.00197",
        #         "close":"0.00197",
        #         "volume":"0.00016676",
        #         "market_volume":"0.08465047"
        #     }
        #
        return [
            self.parse8601(self.safe_string(ohlcv, 'time')),
            self.safe_number(ohlcv, 'open'),
            self.safe_number(ohlcv, 'high'),
            self.safe_number(ohlcv, 'low'),
            self.safe_number(ohlcv, 'close'),
            self.safe_number(ohlcv, 'market_volume'),
        ]

    def fetch_ohlcv(self, symbol, timeframe='5m', since=None, limit=None, params={}):
        """
        fetches historical candlestick data containing the open, high, low, and close price, and the volume of a market
        :param str symbol: unified symbol of the market to fetch OHLCV data for
        :param str timeframe: the length of time each candle represents
        :param int|None since: timestamp in ms of the earliest candle to fetch
        :param int|None limit: the maximum amount of candles to fetch
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns [[int]]: A list of candles ordered as timestamp, open, high, low, close, volume
        """
        self.load_markets()
        market = self.market(symbol)
        request = {
            'market_string': market['id'],
            'interval': self.timeframes[timeframe],
        }
        response = self.publicGetMarketMarketStringOhlcvInterval(self.extend(request, params))
        #
        #     {
        #         "data":{
        #             "slices":[
        #                 {"time":"2019-12-07T22:55:00Z","open":"0.00197","high":"0.00197","low":"0.00197","close":"0.00197","volume":"0.00016676","market_volume":"0.08465047"},
        #                 {"time":"2019-12-07T23:00:00Z","open":"0.00197","high":"0.00197","low":"0.00197","close":"0.00197","volume":"0","market_volume":"0"},
        #                 {"time":"2019-12-07T23:05:00Z","open":"0.00197","high":"0.00197","low":"0.00197","close":"0.00197","volume":"0","market_volume":"0"},
        #             ]
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        ohlcvs = self.safe_value(data, 'slices', [])
        return self.parse_ohlcvs(ohlcvs, market, timeframe, since, limit)

    def fetch_order_book(self, symbol, limit=None, params={}):
        """
        fetches information on open orders with bid(buy) and ask(sell) prices, volumes and other data
        :param str symbol: unified symbol of the market to fetch the order book for
        :param int|None limit: the maximum amount of order book entries to return
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/en/latest/manual.html#order-book-structure>` indexed by market symbols
        """
        self.load_markets()
        market = self.market(symbol)
        request = {'market_string': market['id']}
        response = self.publicGetOrderbookMarketString(self.extend(request, params))
        #
        #     {
        #         "data":{
        #             "buy":{
        #                 "0.00700015":"4.76196367",
        #                 "0.00700017":"1.89755391",
        #                 "0.00700018":"2.13214088",
        #             },
        #             "last_change":1588539869958811,
        #             "sell":{
        #                 "0.02418662":"0.19513696",
        #                 "0.02465627":"0.2439212",
        #                 "0.02530277":"0.663475931274359255",
        #             }
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        orderbook = {}
        sides = {'buy': 'bids', 'sell': 'asks'}
        keys = list(sides.keys())
        for i in range(0, len(keys)):
            key = keys[i]
            side = sides[key]
            bidasks = self.safe_value(data, key, {})
            prices = list(bidasks.keys())
            result = []
            for j in range(0, len(prices)):
                priceAsString = prices[j]
                price = self.safe_number(prices, j)
                amount = self.safe_number(bidasks, priceAsString)
                result.append([price, amount])
            orderbook[side] = result
        timestamp = self.safe_integer_product(data, 'last_change', 0.001)
        return self.parse_order_book(orderbook, market['symbol'], timestamp)

    def parse_ticker(self, ticker, market=None):
        #
        # fetchTicker, fetchTickers
        #
        #     {
        #         "ask":"0.02423119",
        #         "bid":"0.0230939",
        #         "day_avg_price":"0.0247031874349301",
        #         "day_change":"-0.0237543162270376",
        #         "day_high":"0.02470552",
        #         "day_low":"0.02470172",
        #         "day_open":"0.02530277",
        #         "day_volume_base":"0.00268074",
        #         "day_volume_market":"0.10851798",
        #         "id":41,
        #         "id_hr":"ETH_BTC",
        #         "last":"0.02470172",
        #         "last_change":1588533365354609
        #     }
        #
        marketId = self.safe_string(ticker, 'id_hr')
        symbol = self.safe_symbol(marketId, market, '_')
        timestamp = self.safe_integer_product(ticker, 'last_change', 0.001)
        previous = self.safe_string(ticker, 'day_open')
        last = self.safe_string(ticker, 'last')
        day_change = self.safe_string(ticker, 'day_change')
        average = self.safe_string(ticker, 'day_avg_price')
        baseVolume = self.safe_string(ticker, 'day_volume_market')
        quoteVolume = self.safe_string(ticker, 'day_volume_base')
        percentage = Precise.string_mul(day_change, '100')
        change = Precise.string_mul(day_change, previous)
        return self.safe_ticker({
            'symbol': symbol,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'high': self.safe_string(ticker, 'day_high'),
            'low': self.safe_string(ticker, 'day_low'),
            'bid': self.safe_string(ticker, 'bid'),
            'bidVolume': None,
            'ask': self.safe_string(ticker, 'ask'),
            'askVolume': None,
            'vwap': None,
            'open': previous,
            'close': last,
            'last': last,
            'previousClose': None,
            'change': change,
            'percentage': percentage,
            'average': average,
            'baseVolume': baseVolume,
            'quoteVolume': quoteVolume,
            'info': ticker,
        }, market)

    def fetch_tickers(self, symbols=None, params={}):
        """
        fetches price tickers for multiple markets, statistical calculations with the information calculated over the past 24 hours each market
        :param [str]|None symbols: unified symbols of the markets to fetch the ticker for, all market tickers are returned if not assigned
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: an array of `ticker structures <https://docs.ccxt.com/en/latest/manual.html#ticker-structure>`
        """
        self.load_markets()
        symbols = self.market_symbols(symbols)
        response = self.publicGetTickers(params)
        #
        #     {
        #         "data":{
        #             "markets":[
        #                 {
        #                     "ask":"0.0000003",
        #                     "bid":"0.00000029",
        #                     "day_avg_price":"0.0000002999979728",
        #                     "day_change":"0.0344827586206897",
        #                     "day_high":"0.0000003",
        #                     "day_low":"0.0000003",
        #                     "day_open":"0.00000029",
        #                     "day_volume_base":"0.00591958",
        #                     "day_volume_market":"19732.06666665",
        #                     "id":36,
        #                     "id_hr":"DOGE_BTC",
        #                     "last":"0.0000003",
        #                     "last_change":1588534202130778
        #                 },
        #             ]
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        tickers = self.safe_value(data, 'markets', [])
        result = {}
        for i in range(0, len(tickers)):
            ticker = self.parse_ticker(tickers[i])
            symbol = ticker['symbol']
            result[symbol] = ticker
        return self.filter_by_array(result, 'symbol', symbols)

    def fetch_ticker(self, symbol, params={}):
        """
        fetches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific market
        :param str symbol: unified symbol of the market to fetch the ticker for
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: a `ticker structure <https://docs.ccxt.com/en/latest/manual.html#ticker-structure>`
        """
        self.load_markets()
        market = self.market(symbol)
        request = {
            'market_string': market['id'],
        }
        response = self.publicGetTickerMarketString(self.extend(request, params))
        #
        #     {
        #         "data":{
        #             "ask":"0.02423119",
        #             "bid":"0.0230939",
        #             "day_avg_price":"0.0247031874349301",
        #             "day_change":"-0.0237543162270376",
        #             "day_high":"0.02470552",
        #             "day_low":"0.02470172",
        #             "day_open":"0.02530277",
        #             "day_volume_base":"0.00268074",
        #             "day_volume_market":"0.10851798",
        #             "id":41,
        #             "id_hr":"ETH_BTC",
        #             "last":"0.02470172",
        #             "last_change":1588533365354609
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        return self.parse_ticker(data, market)

    def fetch_trades(self, symbol, since=None, limit=None, params={}):
        """
        get the list of most recent trades for a particular symbol
        :param str symbol: unified symbol of the market to fetch trades for
        :param int|None since: timestamp in ms of the earliest trade to fetch
        :param int|None limit: the maximum amount of trades to fetch
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns [dict]: a list of `trade structures <https://docs.ccxt.com/en/latest/manual.html?#public-trades>`
        """
        self.load_markets()
        market = self.market(symbol)
        request = {
            'market_string': market['id'],
            # 'older_than': 123,  # returns trades with id < older_than
            # 'newer_than': 123,  # returns trades with id > newer_than
        }
        response = self.publicGetMarketMarketStringTrades(self.extend(request, params))
        #
        #     {
        #         "data":{
        #             "trades":[
        #                 {
        #                     "id":85507,
        #                     "amount":"0.09390502",
        #                     "price":"0.02556325",
        #                     "base_volume":"0.00240051",
        #                     "seller_taker":true,
        #                     "side":"sell",
        #                     "created_at":"0001-01-01T00:00:00Z",
        #                     "created_at_ts":1581560391338718
        #                 },
        #             ]
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        trades = self.safe_value(data, 'trades', [])
        return self.parse_trades(trades, market, since, limit)

    def fetch_my_trades(self, symbol=None, since=None, limit=None, params={}):
        """
        fetch all trades made by the user
        :param str|None symbol: unified market symbol
        :param int|None since: the earliest time in ms to fetch trades for
        :param int|None limit: the maximum number of trades structures to retrieve
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns [dict]: a list of `trade structures <https://docs.ccxt.com/en/latest/manual.html#trade-structure>`
        """
        self.load_markets()
        request = {
            'desc': True,  # Returns newest trades first when True
            # 'older_than': 123,  # returns trades with id < older_than
            # 'newer_than': 123,  # returns trades with id > newer_than
        }
        market = None
        numericId = self.safe_value(params, 'market_id')
        if numericId is not None:
            request['market_id'] = numericId  # mutually exclusive with market_string
        elif symbol is not None:
            market = self.market(symbol)
            request['market_string'] = market['id']
        response = self.privateGetTrades(self.extend(request, params))
        #
        #     {
        #         "data":{
        #             "trades":[
        #                 {
        #                     "id":107331,
        #                     "market_amount":"0.1082536946986",
        #                     "price":"0.0230939",
        #                     "base_amount":"0.00249999",
        #                     "order_id":13790596,
        #                     "market_id":41,
        #                     "market_string":"ETH_BTC",
        #                     "taker":true,
        #                     "base_fee":"0.00001249",
        #                     "side":"sell",
        #                     "created_at":"2020-05-04T06:08:18.513413Z"
        #                 }
        #             ]
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        trades = self.safe_value(data, 'trades', [])
        return self.parse_trades(trades, market, since, limit)

    def parse_trade(self, trade, market=None):
        #
        # fetchTrades(public)
        #
        #     {
        #         "id":85507,
        #         "amount":"0.09390502",
        #         "price":"0.02556325",
        #         "base_volume":"0.00240051",
        #         "seller_taker":true,
        #         "side":"sell",
        #         "created_at":"0001-01-01T00:00:00Z",
        #         "created_at_ts":1581560391338718
        #     }
        #
        # fetchMyTrades(private)
        #
        #     {
        #         "id":107331,
        #         "market_amount":"0.1082536946986",
        #         "price":"0.0230939",
        #         "base_amount":"0.00249999",
        #         "order_id":13790596,
        #         "market_id":41,
        #         "market_string":"ETH_BTC",
        #         "taker":true,
        #         "base_fee":"0.00001249",
        #         "side":"sell",
        #         "created_at":"2020-05-04T06:08:18.513413Z"
        #     }
        #
        # createOrder, fetchOrders, fetchOpenOrders, fetchClosedOrders
        #
        #     {
        #         "base_amount": "9.58970687",
        #         "base_fee": "0.02397426",
        #         "created_at": "0001-01-01T00:00:00Z",
        #         "id": 0,
        #         "market_amount": "0.97179355",
        #         "price": "9.86804952",
        #         "taker": True
        #     }
        #
        id = self.safe_string(trade, 'id')
        timestamp = self.safe_integer_product(trade, 'created_at_ts', 0.001)
        if timestamp is None:
            timestamp = self.parse8601(self.safe_string(trade, 'created_at'))
        side = self.safe_string(trade, 'side')
        marketId = self.safe_string(trade, 'market_string')
        market = self.safe_market(marketId, market)
        cost = self.safe_string_2(trade, 'base_volume', 'base_amount')
        price = self.safe_string(trade, 'price')
        amount = self.safe_string_2(trade, 'market_amount', 'amount')
        fee = None
        feeCost = self.safe_string(trade, 'base_fee')
        if feeCost is not None:
            feeCurrencyCode = None if (market is None) else market['quote']
            fee = {
                'currency': feeCurrencyCode,
                'cost': feeCost,
            }
        taker = self.safe_value(trade, 'taker', True)
        takerOrMaker = 'taker' if taker else 'maker'
        orderId = self.safe_string(trade, 'order_id')
        return self.safe_trade({
            'id': id,
            'info': trade,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'symbol': market['symbol'],
            'order': orderId,
            'type': None,
            'side': side,
            'takerOrMaker': takerOrMaker,
            'price': price,
            'amount': amount,
            'cost': cost,
            'fee': fee,
        }, market)

    def fetch_trading_fee(self, symbol, params={}):
        """
        fetch the trading fees for a market
        :param str symbol: unified market symbol
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: a `fee structure <https://docs.ccxt.com/en/latest/manual.html#fee-structure>`
        """
        self.load_markets()
        market = self.market(symbol)
        request = {
            'market_string': market['id'],
        }
        response = self.publicGetMarketMarketString(self.extend(request, params))
        #
        #     {
        #         data: {
        #             market: {
        #                 id: '41',
        #                 market_currency: 'ETH',
        #                 base_currency: 'BTC',
        #                 maker_fee: '0',
        #                 taker_fee: '0.005',
        #                 metadata: {},
        #                 can_trade: True,
        #                 can_cancel: True,
        #                 can_view: True,
        #                 market_string: 'ETH_BTC',
        #                 minimum_sell_amount: '0.001',
        #                 minimum_buy_value: '0.0001',
        #                 market_precision: '18',
        #                 base_precision: '8'
        #             },
        #             recent_trades: []
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        marketData = self.safe_value(data, 'market', {})
        return {
            'info': marketData,
            'symbol': market['symbol'],
            'maker': self.safe_number(marketData, 'maker_fee'),
            'taker': self.safe_number(marketData, 'taker_fee'),
            'percentage': True,
            'tierBased': True,
        }

    def parse_balance(self, response):
        data = self.safe_value(response, 'data', {})
        balances = self.safe_value(data, 'balances', [])
        result = {
            'info': response,
            'timestamp': None,
            'datetime': None,
        }
        for i in range(0, len(balances)):
            balance = balances[i]
            currencyId = self.safe_string(balance, 'currency')
            code = self.safe_currency_code(currencyId)
            account = result[code] if (code in result) else self.account()
            account['free'] = self.safe_string(balance, 'balance')
            account['used'] = '0'
            result[code] = account
        balances = self.safe_value(data, 'order_balances', [])
        for i in range(0, len(balances)):
            balance = balances[i]
            currencyId = self.safe_string(balance, 'currency')
            code = self.safe_currency_code(currencyId)
            account = result[code] if (code in result) else self.account()
            account['used'] = self.safe_string(balance, 'balance')
            result[code] = account
        return self.safe_balance(result)

    def fetch_balance(self, params={}):
        """
        query for balance and get the amount of funds available for trading or funds locked in orders
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: a `balance structure <https://docs.ccxt.com/en/latest/manual.html?#balance-structure>`
        """
        self.load_markets()
        response = self.privateGetBalancesAll(params)
        #
        #     {
        #         "data":{
        #             "balances": [
        #                 {"balance": "100000000", "currency": "BCH"},
        #                 {"balance": "99992435.78253015", "currency": "LTC"},
        #                 {"balance": "99927153.76074182", "currency": "BTC"},
        #             ],
        #             "order_balances":[],
        #             "limit_used":0,
        #             "limit_remaining":4000,
        #             "limit":4000
        #         }
        #     }
        #
        return self.parse_balance(response)

    def create_order(self, symbol, type, side, amount, price=None, params={}):
        """
        create a trade order
        :param str symbol: unified symbol of the market to create an order in
        :param str type: 'market' or 'limit'
        :param str side: 'buy' or 'sell'
        :param float amount: how much of currency you want to trade in units of base currency
        :param float|None price: the price at which the order is to be fullfilled, in units of the quote currency, ignored in market orders
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: an `order structure <https://docs.ccxt.com/en/latest/manual.html#order-structure>`
        """
        if type != 'limit':
            raise InvalidOrder(self.id + ' createOrder() allows limit orders only')
        self.load_markets()
        market = self.market(symbol)
        request = {
            'amount': self.amount_to_precision(market['symbol'], amount),
            'market_id': market['numericId'],
            'price': self.price_to_precision(market['symbol'], price),
        }
        method = 'privatePostSellLimit' if (side == 'sell') else 'privatePostBuyLimit'
        response = getattr(self, method)(self.extend(request, params))
        #
        #     {
        #         "data": {
        #             "order": {
        #                 "created_at": "2018-04-06T20:46:52.899248Z",
        #                 "id": 13253,
        #                 "market_amount": "1",
        #                 "market_amount_remaining": "0",
        #                 "market_id": 1,
        #                 "open": False,
        #                 "order_type": "sell_limit",
        #                 "price": "0.01",
        #                 "trades": [
        #                     {
        #                         "base_amount": "0.27834267",
        #                         "base_fee": "0.00069585",
        #                         "created_at": "0001-01-01T00:00:00Z",
        #                         "id": 0,
        #                         "market_amount": "0.02820645",
        #                         "price": "9.86805058",
        #                         "taker": True
        #                     },
        #                     {
        #                         "base_amount": "9.58970687",
        #                         "base_fee": "0.02397426",
        #                         "created_at": "0001-01-01T00:00:00Z",
        #                         "id": 0,
        #                         "market_amount": "0.97179355",
        #                         "price": "9.86804952",
        #                         "taker": True
        #                     }
        #                 ]
        #             }
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        order = self.safe_value(data, 'order', {})
        return self.parse_order(order, market)

    def parse_order(self, order, market=None):
        #
        # createOrder
        #
        #     {
        #         "created_at": "2018-04-06T20:46:52.899248Z",
        #         "id": 13253,
        #         "market_amount": "1",
        #         "market_amount_remaining": "0",
        #         "market_id": 1,
        #         "open": False,
        #         "order_type": "sell_limit",
        #         "price": "0.01",
        #         "trades": [
        #             {
        #                 "base_amount": "0.27834267",
        #                 "base_fee": "0.00069585",
        #                 "created_at": "0001-01-01T00:00:00Z",
        #                 "id": 0,
        #                 "market_amount": "0.02820645",
        #                 "price": "9.86805058",
        #                 "taker": True
        #             },
        #             {
        #                 "base_amount": "9.58970687",
        #                 "base_fee": "0.02397426",
        #                 "created_at": "0001-01-01T00:00:00Z",
        #                 "id": 0,
        #                 "market_amount": "0.97179355",
        #                 "price": "9.86804952",
        #                 "taker": True
        #             }
        #         ]
        #     }
        #
        # fetchOrder
        #
        #     {
        #         id: 13790596,
        #         market_amount: "0.15",
        #         market_amount_remaining: "0",
        #         created_at: "2020-05-04T06:08:18.513413Z",
        #         price: "0.0230939",
        #         base_amount: "0",
        #         order_type: "sell_limit",
        #         market_id: 41,
        #         market_string: "ETH_BTC",
        #         open: False,
        #         trades: [
        #             {
        #                 id: 107331,
        #                 market_amount: "0.1082536946986",
        #                 price: "0.0230939",
        #                 base_amount: "0.00249999",
        #                 taker: True,
        #                 base_fee: "0.00001249",
        #                 created_at: "2020-05-04T06:08:18.513413Z",
        #             }
        #         ],
        #         close_reason: "canceled"
        #     }
        #
        id = self.safe_string(order, 'id')
        timestamp = self.parse8601(self.safe_string(order, 'created_at'))
        sideType = self.safe_string(order, 'order_type')
        orderType = None
        side = None
        if sideType is not None:
            parts = sideType.split('_')
            side = self.safe_string(parts, 0)
            orderType = self.safe_string(parts, 1)
        price = self.safe_string(order, 'price')
        amount = self.safe_string(order, 'market_amount')
        remaining = self.safe_string(order, 'market_amount_remaining')
        open = self.safe_value(order, 'open', False)
        closeReason = self.safe_string(order, 'close_reason')
        status = None
        if open:
            status = 'open'
        elif closeReason == 'canceled':
            status = 'canceled'
        else:
            status = 'closed'
        marketId = self.safe_string(order, 'market_string')
        market = self.safe_market(marketId, market, '_')
        symbol = market['symbol']
        rawTrades = self.safe_value(order, 'trades', [])
        return self.safe_order({
            'info': order,
            'id': id,
            'clientOrderId': None,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'lastTradeTimestamp': None,
            'symbol': symbol,
            'type': orderType,
            'timeInForce': None,
            'postOnly': None,
            'side': side,
            'price': price,
            'stopPrice': None,
            'average': None,
            'amount': amount,
            'remaining': remaining,
            'filled': None,
            'status': status,
            'fee': None,
            'fees': None,
            'cost': None,
            'trades': rawTrades,
        }, market)

    def cancel_order(self, id, symbol=None, params={}):
        """
        cancels an open order
        :param str id: order id
        :param str|None symbol: not used by qtrade cancelOrder()
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: An `order structure <https://docs.ccxt.com/en/latest/manual.html#order-structure>`
        """
        request = {
            'id': int(id),
        }
        # successful cancellation returns 200 with no payload
        return self.privatePostCancelOrder(self.extend(request, params))

    def fetch_order(self, id, symbol=None, params={}):
        """
        fetches information on an order made by the user
        :param str|None symbol: not used by qtrade fetchOrder
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: An `order structure <https://docs.ccxt.com/en/latest/manual.html#order-structure>`
        """
        self.load_markets()
        request = {'order_id': id}
        response = self.privateGetOrderOrderId(self.extend(request, params))
        #
        #     {
        #         "data":{
        #             "order":{
        #                 "id":13790596,
        #                 "market_amount":"0.15",
        #                 "market_amount_remaining":"0.0417463053014",
        #                 "created_at":"2020-05-04T06:08:18.513413Z",
        #                 "price":"0.0230939",
        #                 "order_type":"sell_limit",
        #                 "market_id":41,
        #                 "market_string":"ETH_BTC",
        #                 "open":true,
        #                 "trades":[
        #                     {
        #                         "id":107331,
        #                         "market_amount":"0.1082536946986",
        #                         "price":"0.0230939",
        #                         "base_amount":"0.00249999",
        #                         "taker":true,
        #                         "base_fee":"0.00001249",
        #                         "created_at":"2020-05-04T06:08:18.513413Z"
        #                     }
        #                 ]
        #             }
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        order = self.safe_value(data, 'order', {})
        return self.parse_order(order)

    def fetch_orders(self, symbol=None, since=None, limit=None, params={}):
        """
        fetches information on multiple orders made by the user
        :param str|None symbol: unified market symbol of the market orders were made in
        :param int|None since: the earliest time in ms to fetch orders for
        :param int|None limit: the maximum number of  orde structures to retrieve
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns [dict]: a list of `order structures <https://docs.ccxt.com/en/latest/manual.html#order-structure>`
        """
        self.load_markets()
        request = {
            # 'open': True,
            # 'older_than': 123,  # returns orders with id < older_than
            # 'newer_than': 123,  # returns orders with id > newer_than
        }
        market = None
        numericId = self.safe_value(params, 'market_id')
        if numericId is not None:
            request['market_id'] = numericId  # mutually exclusive with market_string
        elif symbol is not None:
            market = self.market(symbol)
            request['market_string'] = market['id']
        response = self.privateGetOrders(self.extend(request, params))
        #
        #     {
        #         "data":{
        #             "orders":[
        #                 {
        #                     "id":13790596,
        #                     "market_amount":"0.15",
        #                     "market_amount_remaining":"0.0417463053014",
        #                     "created_at":"2020-05-04T06:08:18.513413Z",
        #                     "price":"0.0230939",
        #                     "order_type":"sell_limit",
        #                     "market_id":41,
        #                     "market_string":"ETH_BTC",
        #                     "open":true,
        #                     "trades":[
        #                         {
        #                             "id":107331,
        #                             "market_amount":"0.1082536946986",
        #                             "price":"0.0230939",
        #                             "base_amount":"0.00249999",
        #                             "taker":true,
        #                             "base_fee":"0.00001249",
        #                             "created_at":"2020-05-04T06:08:18.513413Z"
        #                         }
        #                     ]
        #                 }
        #             ]
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        orders = self.safe_value(data, 'orders', [])
        return self.parse_orders(orders, market, since, limit)

    def fetch_open_orders(self, symbol=None, since=None, limit=None, params={}):
        """
        fetch all unfilled currently open orders
        :param str|None symbol: unified market symbol
        :param int|None since: the earliest time in ms to fetch open orders for
        :param int|None limit: the maximum number of  open orders structures to retrieve
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns [dict]: a list of `order structures <https://docs.ccxt.com/en/latest/manual.html#order-structure>`
        """
        request = {'open': True}
        return self.fetch_orders(symbol, since, limit, self.extend(request, params))

    def fetch_closed_orders(self, symbol=None, since=None, limit=None, params={}):
        """
        fetches information on multiple closed orders made by the user
        :param str|None symbol: unified market symbol of the market orders were made in
        :param int|None since: the earliest time in ms to fetch orders for
        :param int|None limit: the maximum number of  orde structures to retrieve
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns [dict]: a list of `order structures <https://docs.ccxt.com/en/latest/manual.html#order-structure>`
        """
        request = {'open': False}
        return self.fetch_orders(symbol, since, limit, self.extend(request, params))

    def parse_deposit_address(self, depositAddress, currency=None):
        #
        #     {
        #         "address":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #         "currency_status":"ok",
        #         "deposit_methods":{
        #             "address":{
        #                 "deposit_type":"address",
        #                 "render_type":"address",
        #                 "label":"Address",
        #                 "address":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #             },
        #         },
        #     }
        #
        code = None if (currency is None) else currency['code']
        address = self.safe_string(depositAddress, 'address')
        tag = None
        if address is not None:
            parts = address.split(':')
            address = self.safe_string(parts, 0)
            tag = self.safe_string(parts, 1)
        self.check_address(address)
        return {
            'currency': code,
            'address': address,
            'tag': tag,
            'network': None,
            'info': depositAddress,
        }

    def fetch_deposit_address(self, code, params={}):
        """
        fetch the deposit address for a currency associated with self account
        :param str code: unified currency code
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: an `address structure <https://docs.ccxt.com/en/latest/manual.html#address-structure>`
        """
        self.load_markets()
        currency = self.currency(code)
        request = {
            'currency': currency['id'],
        }
        response = self.privatePostDepositAddressCurrency(self.extend(request, params))
        #
        #     {
        #         "data":{
        #             "address":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #             "currency_status":"ok",
        #             "deposit_methods":{
        #                 "address":{
        #                     "deposit_type":"address",
        #                     "render_type":"address",
        #                     "label":"Address",
        #                     "address":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #                 },
        #             },
        #         },
        #     }
        #
        data = self.safe_value(response, 'data', {})
        return self.parse_deposit_address(data, currency)

    def fetch_deposit(self, id, code=None, params={}):
        """
        fetch information on a deposit
        :param str id: deposit id
        :param str|None code: not used by qtrade fetchDeposit()
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: a `transaction structure <https://docs.ccxt.com/en/latest/manual.html#transaction-structure>`
        """
        self.load_markets()
        request = {
            'deposit_id': id,
        }
        response = self.privateGetDepositDepositId(self.extend(request, params))
        #
        #     {
        #         "data":{
        #             "deposit":{
        #                 "id":"0xaa6e65ed274c4786e5dec3671de96f81021cacdbc453b1a133ab84356f3620a0",
        #                 "amount":"0.13",
        #                 "currency":"ETH",
        #                 "address":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #                 "status":"credited",
        #                 "relay_status":"",
        #                 "network_data":{
        #                     "confirms":87,
        #                     "sweep_txid":"0xa16e65ed274d4686e5dec3671de96f81021cacdbc453b1a133ab85356f3630a0",
        #                     "sweep_balance":"0.150000000000000000",
        #                     "confirms_required":80,
        #                     "unsigned_sweep_tx":{
        #                         "chainId":1,
        #                         "from":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #                         "gas":"0x5208",
        #                         "gasPrice":"0x19b45a500",
        #                         "nonce":"0x0",
        #                         "to":"0x76Cd80202a2C31e9D8F595a31ed071CE7F75BB93",
        #                         "value":"0x214646b6347d800"
        #                     },
        #                     "txid":"0xaa6e65ed274c4786e5dec3671de96f81021cacdbc453b1a133ab84356f3620a0",
        #                     "tx_index":"0x6f",
        #                     "tx_value":"0.130000000000000000",
        #                     "key_index":311,
        #                     "blockheight":9877869,
        #                     "signed_sweep_tx":{
        #                         "hash":"0xa16e65ed274d4686e5dec3671de96f81021cacdbc453b1a133ab85356f3630a0",
        #                         "rawTransaction":"0xd86c8085019b45a1008252099476cb80202b2c31e9d7f595a31fd071ce7f75bb93880214646b6347d8008046a08c6e3bfe8b25bff2b6851c87ea17c63d7b23591210ab0779a568eaa43dc40435a030e964bb2b667072ea7cbc8ab554403e3f3ead9b554743f2fdc2b1e06e998df9"
        #                     },
        #                     "estimated_sweep_tx_fee":144900000000000
        #                 },
        #                 "created_at":"2020-05-04T05:38:42.145162Z"
        #             }
        #         }
        #     }
        data = self.safe_value(response, 'data', {})
        deposit = self.safe_value(data, 'deposit', {})
        return self.parse_transaction(deposit)

    def fetch_deposits(self, code=None, since=None, limit=None, params={}):
        """
        fetch all deposits made to an account
        :param str|None code: unified currency code
        :param int|None since: the earliest time in ms to fetch deposits for
        :param int|None limit: the maximum number of deposits structures to retrieve
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns [dict]: a list of `transaction structures <https://docs.ccxt.com/en/latest/manual.html#transaction-structure>`
        """
        self.load_markets()
        currency = None
        if code is not None:
            currency = self.currency(code)
        response = self.privateGetDeposits(params)
        #
        #     {
        #         "data":{
        #             "deposits":[
        #                 {
        #                     "id":"0xaa6e65ed274c4786e5dec3671de96f81021cacdbc453b1a133ab84356f3620a0",
        #                     "amount":"0.13",
        #                     "currency":"ETH",
        #                     "address":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #                     "status":"credited",
        #                     "relay_status":"",
        #                     "network_data":{
        #                         "confirms":87,
        #                         "sweep_txid":"0xa16e65ed274d4686e5dec3671de96f81021cacdbc453b1a133ab85356f3630a0",
        #                         "sweep_balance":"0.150000000000000000",
        #                         "confirms_required":80,
        #                         "unsigned_sweep_tx":{
        #                             "chainId":1,
        #                             "from":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #                             "gas":"0x5208",
        #                             "gasPrice":"0x19b45a500",
        #                             "nonce":"0x0",
        #                             "to":"0x76Cd80202a2C31e9D8F595a31ed071CE7F75BB93",
        #                             "value":"0x214646b6347d800"
        #                         },
        #                         "txid":"0xaa6e65ed274c4786e5dec3671de96f81021cacdbc453b1a133ab84356f3620a0",
        #                         "tx_index":"0x6f",
        #                         "tx_value":"0.130000000000000000",
        #                         "key_index":311,
        #                         "blockheight":9877869,
        #                         "signed_sweep_tx":{
        #                             "hash":"0xa16e65ed274d4686e5dec3671de96f81021cacdbc453b1a133ab85356f3630a0",
        #                             "rawTransaction":"0xd86c8085019b45a1008252099476cb80202b2c31e9d7f595a31fd071ce7f75bb93880214646b6347d8008046a08c6e3bfe8b25bff2b6851c87ea17c63d7b23591210ab0779a568eaa43dc40435a030e964bb2b667072ea7cbc8ab554403e3f3ead9b554743f2fdc2b1e06e998df9"
        #                         },
        #                         "estimated_sweep_tx_fee":144900000000000
        #                     },
        #                     "created_at":"2020-05-04T05:38:42.145162Z"
        #                 }
        #             ]
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        deposits = self.safe_value(data, 'deposits', [])
        return self.parse_transactions(deposits, currency, since, limit)

    def fetch_transfers(self, code=None, since=None, limit=None, params={}):
        """
        fetch a history of internal transfers made on an account
        :param str|None code: unified currency code of the currency transferred
        :param int|None since: the earliest time in ms to fetch transfers for
        :param int|None limit: the maximum number of  transfers structures to retrieve
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns [dict]: a list of `transfer structures <https://docs.ccxt.com/en/latest/manual.html#transfer-structure>`
        """
        self.load_markets()
        currency = None
        if code is not None:
            currency = self.currency(code)
        response = self.privateGetTransfers(params)
        #
        #     {
        #         "data": {
        #             "transfers": [
        #                 {
        #                     "amount": "0.5",
        #                     "created_at": "2018-12-10T00:06:41.066665Z",
        #                     "currency": "BTC",
        #                     "id": 9,
        #                     "reason_code": "referral_payout",
        #                     "reason_metadata": {
        #                         "note": "January referral earnings"
        #                     },
        #                     "sender_email": "qtrade",
        #                     "sender_id": 218
        #                 }
        #             ]
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        transfers = self.safe_value(data, 'transfers', [])
        return self.parse_transfers(transfers, currency, since, limit)

    def parse_transfer(self, transfer, currency=None):
        #
        #     {
        #         "amount": "0.5",
        #         "created_at": "2018-12-10T00:06:41.066665Z",
        #         "currency": "BTC",
        #         "id": 9,
        #         "reason_code": "referral_payout",
        #         "reason_metadata": {
        #             "note": "January referral earnings"
        #         },
        #         "sender_email": "qtrade",
        #         "sender_id": 218
        #     }
        #
        currencyId = self.safe_string(transfer, 'currency')
        dateTime = self.safe_string(transfer, 'created_at')
        return {
            'info': transfer,
            'id': self.safe_string(transfer, 'id'),
            'timestamp': self.parse8601(dateTime),
            'datetime': dateTime,
            'currency': self.safe_currency_code(currencyId, currency),
            'amount': self.safe_number(transfer, 'amount'),
            'fromAccount': self.safe_string(transfer, 'sender_id'),
            'toAccount': None,
            'status': 'ok',
        }

    def fetch_withdrawal(self, id, code=None, params={}):
        """
        fetch data on a currency withdrawal via the withdrawal id
        :param str id: withdrawal id
        :param str|None code: not used by qtrade.fetchWithdrawal
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: a `transaction structure <https://docs.ccxt.com/en/latest/manual.html#transaction-structure>`
        """
        self.load_markets()
        request = {
            'withdraw_id': id,
        }
        response = self.privateGetWithdrawWithdrawId(self.extend(request, params))
        #
        #     {
        #         data: {
        #             withdraw: {
        #                 "id":25524,
        #                 "amount":"0.0417463053014",
        #                 "user_id":0,
        #                 "currency":"ETH",
        #                 "network_data":{
        #                     "unsigned_tx":{
        #                         "chainId":1,
        #                         "from":"0x76Cd80202a2C31e9D8F595a31ed071CE7F75BB93",
        #                         "gas":"0x5208",
        #                         "gasPrice":"0x20c8558e9",
        #                         "nonce":"0xf3",
        #                         "to":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #                         "value":"0x71712bcd113308"
        #                     },
        #                     "estimated_tx_fee":184800004893000,
        #                     "confirms_required":80,
        #                     "txid":"0x79439b62473d61d99ce1dc6c3b8a417da36d45323a394bb0d4af870608fef38d",
        #                     "confirms":83,
        #                     "signed_tx":{
        #                         "hash":"0x79439b62473d61d99ce1dc6c3b8a417da36d45323a394bb0d4af870608fef38d",
        #                         "rawTransaction":"0xf86c81f385021c8558e98252089401b0a9b7b4cde774af0f3e87cb4f1c2ccdba08068771712acd1133078025a0088157d119d924d47413c81b91b9f18ff148623a2ef13dab1895ca3ba546b771a046a021b1e1f64d1a60bb66c19231f641b352326188a9ed3b931b698a939f78d0"
        #                     }
        #                 },
        #                 "address":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #                 "status":"confirmed",
        #                 "relay_status":"",
        #                 "created_at":"2020-05-05T06:32:19.907061Z",
        #                 "cancel_requested":false
        #             }
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        withdrawal = self.safe_value(data, 'withdraw', {})
        return self.parse_transaction(withdrawal)

    def fetch_withdrawals(self, code=None, since=None, limit=None, params={}):
        """
        fetch all withdrawals made from an account
        :param str|None code: unified currency code
        :param int|None since: the earliest time in ms to fetch withdrawals for
        :param int|None limit: the maximum number of withdrawals structures to retrieve
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns [dict]: a list of `transaction structures <https://docs.ccxt.com/en/latest/manual.html#transaction-structure>`
        """
        self.load_markets()
        currency = None
        if code is not None:
            currency = self.currency(code)
        response = self.privateGetWithdraws(params)
        #     {
        #         "data":{
        #             "withdraws":[
        #                 {
        #                     "id":25524,
        #                     "amount":"0.0417463053014",
        #                     "user_id":0,
        #                     "currency":"ETH",
        #                     "network_data":{
        #                         "unsigned_tx":{
        #                             "chainId":1,
        #                             "from":"0x76Cd80202a2C31e9D8F595a31ed071CE7F75BB93",
        #                             "gas":"0x5208",
        #                             "gasPrice":"0x20c8558e9",
        #                             "nonce":"0xf3",
        #                             "to":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #                             "value":"0x71712bcd113308"
        #                         },
        #                         "estimated_tx_fee":184800004893000,
        #                         "confirms_required":80,
        #                         "txid":"0x79439b62473d61d99ce1dc6c3b8a417da36d45323a394bb0d4af870608fef38d",
        #                         "confirms":83,
        #                         "signed_tx":{
        #                             "hash":"0x79439b62473d61d99ce1dc6c3b8a417da36d45323a394bb0d4af870608fef38d",
        #                             "rawTransaction":"0xf86c81f385021c8558e98252089401b0a9b7b4cde774af0f3e87cb4f1c2ccdba08068771712acd1133078025a0088157d119d924d47413c81b91b9f18ff148623a2ef13dab1895ca3ba546b771a046a021b1e1f64d1a60bb66c19231f641b352326188a9ed3b931b698a939f78d0"
        #                         }
        #                     },
        #                     "address":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #                     "status":"confirmed",
        #                     "relay_status":"",
        #                     "created_at":"2020-05-05T06:32:19.907061Z",
        #                     "cancel_requested":false
        #                 }
        #             ]
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        withdrawals = self.safe_value(data, 'withdraws', [])
        return self.parse_transactions(withdrawals, currency, since, limit)

    def parse_transaction(self, transaction, currency=None):
        #
        # fetchDeposits, fetchDeposit
        #
        #     {
        #         "id":"0xaa6e65ed274c4786e5dec3671de96f81021cacdbc453b1a133ab84356f3620a0",
        #         "amount":"0.13",
        #         "currency":"ETH",
        #         "address":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #         "status":"credited",
        #         "relay_status":"",
        #         "network_data":{
        #             "confirms":87,
        #             "sweep_txid":"0xa16e65ed274d4686e5dec3671de96f81021cacdbc453b1a133ab85356f3630a0",
        #             "sweep_balance":"0.150000000000000000",
        #             "confirms_required":80,
        #             "unsigned_sweep_tx":{
        #                 "chainId":1,
        #                 "from":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #                 "gas":"0x5208",
        #                 "gasPrice":"0x19b45a500",
        #                 "nonce":"0x0",
        #                 "to":"0x76Cd80202a2C31e9D8F595a31ed071CE7F75BB93",
        #                 "value":"0x214646b6347d800"
        #             },
        #             "txid":"0xaa6e65ed274c4786e5dec3671de96f81021cacdbc453b1a133ab84356f3620a0",
        #             "tx_index":"0x6f",
        #             "tx_value":"0.130000000000000000",
        #             "key_index":311,
        #             "blockheight":9877869,
        #             "signed_sweep_tx":{
        #                 "hash":"0xa16e65ed274d4686e5dec3671de96f81021cacdbc453b1a133ab85356f3630a0",
        #                 "rawTransaction":"0xd86c8085019b45a1008252099476cb80202b2c31e9d7f595a31fd071ce7f75bb93880214646b6347d8008046a08c6e3bfe8b25bff2b6851c87ea17c63d7b23591210ab0779a568eaa43dc40435a030e964bb2b667072ea7cbc8ab554403e3f3ead9b554743f2fdc2b1e06e998df9"
        #             },
        #             "estimated_sweep_tx_fee":144900000000000
        #         },
        #         "created_at":"2020-05-04T05:38:42.145162Z"
        #     }
        #
        # fetchWithdrawals, fetchWithdrawal
        #
        #     {
        #         "id":25524,
        #         "amount":"0.0417463053014",
        #         "user_id":0,
        #         "currency":"ETH",
        #         "network_data":{
        #             "unsigned_tx":{
        #                 "chainId":1,
        #                 "from":"0x76Cd80202a2C31e9D8F595a31ed071CE7F75BB93",
        #                 "gas":"0x5208",
        #                 "gasPrice":"0x20c8558e9",
        #                 "nonce":"0xf3",
        #                 "to":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #                 "value":"0x71712bcd113308"
        #             },
        #             "estimated_tx_fee":184800004893000,
        #             "confirms_required":80,
        #             "txid":"0x79439b62473d61d99ce1dc6c3b8a417da36d45323a394bb0d4af870608fef38d",
        #             "confirms":83,
        #             "signed_tx":{
        #                 "hash":"0x79439b62473d61d99ce1dc6c3b8a417da36d45323a394bb0d4af870608fef38d",
        #                 "rawTransaction":"0xf86c81f385021c8558e98252089401b0a9b7b4cde774af0f3e87cb4f1c2ccdba08068771712acd1133078025a0088157d119d924d47413c81b91b9f18ff148623a2ef13dab1895ca3ba546b771a046a021b1e1f64d1a60bb66c19231f641b352326188a9ed3b931b698a939f78d0"
        #             }
        #         },
        #         "address":"0xe0cd26f9A60118555247aE6769A5d241D91f07f2",
        #         "status":"confirmed",
        #         "relay_status":"",
        #         "created_at":"2020-05-05T06:32:19.907061Z",
        #         "cancel_requested":false
        #     }
        #
        # withdraw
        #
        #     {
        #         "code": "initiated",
        #         "id": 3,
        #         "result": "Withdraw initiated. Please allow 3-5 minutes for our system to process."
        #     }
        #
        timestamp = self.parse8601(self.safe_string(transaction, 'created_at'))
        id = self.safe_string(transaction, 'id')
        networkData = self.safe_value(transaction, 'network_data', {})
        unsignedTx = self.safe_value(networkData, 'unsigned_tx', {})
        addressFrom = self.safe_string(unsignedTx, 'from')
        txid = self.safe_string(networkData, 'txid')
        address = self.safe_string(transaction, 'address')
        tag = None
        if address is not None:
            parts = address.split(':')
            numParts = len(parts)
            if numParts > 1:
                address = self.safe_string(parts, 0)
                tag = self.safe_string(parts, 1)
        addressTo = address
        tagFrom = None
        tagTo = tag
        cancelRequested = self.safe_value(transaction, 'cancel_requested')
        type = 'deposit' if (cancelRequested is None) else 'withdrawal'
        amount = self.safe_number(transaction, 'amount')
        currencyId = self.safe_string(transaction, 'currency')
        code = self.safe_currency_code(currencyId)
        status = self.parse_transaction_status(self.safe_string(transaction, 'status'))
        statusCode = self.safe_string(transaction, 'code')
        if cancelRequested:
            status = 'canceled'
        elif status is None:
            status = self.parse_transaction_status(statusCode)
        fee = None
        return {
            'info': transaction,
            'id': id,
            'txid': txid,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'network': None,
            'addressFrom': addressFrom,
            'addressTo': addressTo,
            'address': address,
            'tagFrom': tagFrom,
            'tagTo': tagTo,
            'tag': tag,
            'type': type,
            'amount': amount,
            'currency': code,
            'status': status,
            'updated': None,
            'fee': fee,
        }

    def parse_transaction_status(self, status):
        statuses = {
            'initiated': 'pending',
            'needs_create': 'pending',
            'credited': 'ok',
            'confirmed': 'ok',
        }
        return self.safe_string(statuses, status, status)

    def withdraw(self, code, amount, address, tag=None, params={}):
        """
        make a withdrawal
        :param str code: unified currency code
        :param float amount: the amount to withdraw
        :param str address: the address to withdraw to
        :param str|None tag:
        :param dict params: extra parameters specific to the qtrade api endpoint
        :returns dict: a `transaction structure <https://docs.ccxt.com/en/latest/manual.html#transaction-structure>`
        """
        tag, params = self.handle_withdraw_tag_and_params(tag, params)
        self.load_markets()
        currency = self.currency(code)
        request = {
            'address': address,
            'amount': amount,
            'currency': currency['id'],
        }
        if tag is not None:
            request['address'] += ':' + tag
        response = self.privatePostWithdraw(self.extend(request, params))
        #
        #     {
        #         "data": {
        #             "code": "initiated",
        #             "id": 3,
        #             "result": "Withdraw initiated. Please allow 3-5 minutes for our system to process."
        #         }
        #     }
        #
        data = self.safe_value(response, 'data', {})
        result = self.parse_transaction(data)
        return self.extend(result, {
            'currency': code,
            'address': address,
            'addressTo': address,
            'tag': tag,
            'tagTo': tag,
            'amount': amount,
        })

    def nonce(self):
        return self.milliseconds()

    def sign(self, path, api='public', method='GET', params={}, headers=None, body=None):
        url = '/' + self.version + '/'
        if api == 'private':
            url += 'user/'
        url += self.implode_params(path, params)
        request = self.omit(params, self.extract_params(path))
        if method == 'POST':
            body = self.json(request)
        else:
            if request:
                url += '?' + self.urlencode(request)
        if api == 'private':
            timestamp = str(self.milliseconds())
            bodyAsString = body if (method == 'POST') else ''
            auth = "\n".join([
                method,
                url,
                timestamp,
                bodyAsString,
                self.secret,
            ])  # eslint-disable-line quotes
            hash = self.hash(self.encode(auth), 'sha256', 'base64')
            key = self.apiKey
            if not isinstance(key, str):
                key = str(key)
            signature = 'HMAC-SHA256 ' + key + ':' + hash
            headers = {
                'Authorization': signature,
                'HMAC-Timestamp': timestamp,
            }
            if method == 'POST':
                headers['Content-Type'] = 'application/json'
        url = self.urls['api']['rest'] + url
        return {'url': url, 'method': method, 'body': body, 'headers': headers}

    def handle_errors(self, code, reason, url, method, headers, body, response, requestHeaders, requestBody):
        #
        #     {"errors":[{"code":"insuff_funds","title":"Your available balance is too low for that action"}]}
        #     {"errors":[{"code": "invalid_auth","title": "Invalid HMAC signature"}]}
        #
        if response is None:
            return
        errors = self.safe_value(response, 'errors', [])
        numErrors = len(errors)
        if numErrors < 1:
            return
        feedback = self.id + ' ' + body
        for i in range(0, len(errors)):
            error = errors[i]
            errorCode = self.safe_string(error, 'code')
            self.throw_exactly_matched_exception(self.exceptions['exact'], errorCode, feedback)
        raise ExchangeError(feedback)  # unknown message
